#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/syscall.h>
#include <sys/types.h>
#include <stdarg.h>
#include <sys/fcntl.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/uio.h> /* for readv */
#include <sys/socket.h>
#include <errno.h>
#include <sys/sysctl.h>
#include <sys/time.h> /* for setitimer */
#include <signal.h>
#include <sys/dirent.h> /* for getdirentries */

#define DC_INTERNAL
#include "dc.h"

#ifdef TWO_PHASE_COMMIT
#include "include/2pc.h"
#endif

#if !defined(TRACK_ND) && !defined(TRACK_NA)
#   error "must define one of TRACK_ND or TRACK_NA"
#endif

/*
 * global variables
 */
vista_segment *data_seg_ptr, *explicit_seg_ptr, *dc_seg_ptr=NULL;
explicit_type *explicit_ptr = NULL;
char dc_stack[MAX_STACK_SIZE];
char *dc_stack_public = dc_stack;
struct sigaction dc_siglist[NSIG];	/* which signals are being handled by
					    the user */
void (*user_recov_ptr)() = NULL;	/* user-specified recovery function */
void (*user_check_ptr)() = NULL;	/* user-specified checkpoint function */
char dc_command[200];			/* command tag for dc files */
struct itimerval timer;
struct sockaddr_in serv_addr;
int socket_2pc;				/* receives messages from 2PC coord. */
pid_t original_pid;			/* original pid for this process */
pid_t original_ppid;			/* original pid for the parent of this
					    process */
struct timeval t_start;			/* starting time of the program */

/* ??? */
char stdout_buf[1024];

#ifdef DC_DISK
#define USER_FD 13			/* first file descriptor that can be
					    assigned to a user open.  Everything
					    below this number is used by
					    Vista or DC */
#else
#define USER_FD 10
#endif

extern void vista_sigprot();
extern void vista_sigother();

/*
 * DC's lock routines.  Prevent signals from being handled, but allow
 * prot faults.  Implemented with vista's locking primitives.
 */
dc_lock()
{
    /*
     * Prefault so that vista_lock and prot_unlock don't cause a prot fault
     * with the prot_lock held.  I think there's a race condition here
     * if we get a signal between prefault() and vista_lock().
     */
    prefault();
    vista_lock();
    prot_unlock();
}

dc_unlock()
{
    /*
     * Prefault so that prot_lock and vista_unlock don't cause a prot fault
     * with the prot_lock held.
     */
    prefault();
    prot_lock();
    vista_unlock();
}

/*
 * Make sure a buffer is not write protected, so that system calls (e.g.
 * read won't return EFAULT).
 */
dc_touch(char *start, int length)
{
    int i;
#ifdef APP_MPROTECT
    extern char app_protect[NUMPAGES];
#endif

    /* fprintf(stderr, "dc_touch 0x%x (%d)\t", start, length); */

    if (start == NULL) {
	return;
    }

    for (i = 0; i < length; i += VISTA_PAGE) {
#ifdef APP_MPROTECT
	if (!(app_protect[(unsigned long) (start+i) / VISTA_PAGE] & PROT_WRITE)) {
	    fprintf(stderr, "\t0x%lx protected *****\n", start+i);
	    exit(100);
	}
#endif
	*(start+i) = *(start+i);
    }
    if (length) {
#ifdef APP_MPROTECT
	if (! (app_protect[(unsigned long) (start+length-1) / VISTA_PAGE] &
		PROT_WRITE)) {
	    fprintf(stderr, "\t0x%lx protected *****\n", start+length-1);
	    exit(100);
	    return;
	}
#endif
	*(start+length-1) = *(start+length-1);
    }
}

int socket(int domain, int type, int protocol)
{
    int sock;

    sock = syscall(SYS_socket, domain, type, protocol);

    if (domain != AF_INET || type != SOCK_DGRAM) {
	fprintf(stderr, "***********************************************\n");
	fprintf(stderr, "warning: socket %d called with domain = %d, type = %d\n",
	    sock, domain, type);
	fprintf(stderr, "socket must be manually reconstructed at recovery\n");
	fprintf(stderr, "***********************************************\n");
	return(sock);
    }

    if (sock >= USER_FD) {
	prefault();
	vista_set_range(dc_seg_ptr, &(explicit_ptr->fds[sock]),
			sizeof(explicit_ptr->fds[sock]), explicit_ptr->dc_tid);
	explicit_ptr->fds[sock].status = FD_SOCKET;
	explicit_ptr->fds[sock].select_buf = NULL;
	explicit_ptr->fds[sock].u.sock.domain = domain;
	explicit_ptr->fds[sock].u.sock.type = type;
	explicit_ptr->fds[sock].u.sock.protocol = protocol;
	explicit_ptr->fds[sock].u.sock.bound = 0;
	explicit_ptr->fds[sock].u.sock.connected = 0;
	explicit_ptr->fds[sock].u.sock.sigio = 0;
	explicit_ptr->fds[sock].u.sock.nonblock = 0;
	explicit_ptr->fds[sock].u.sock.sndbuf = 0;
	explicit_ptr->fds[sock].u.sock.rcvbuf = 0;
	explicit_ptr->fds[sock].u.sock.broadcast = 0;

    }

    return(sock);
}

/*
 * Trap setsockopt to recover SO_SNDBUF and SO_RCVBUF.
 */
int setsockopt(int s, int level, int optname, const void *optval, int optlen)
{
    int i;

    if (explicit_ptr->fds[s].status == FD_SOCKET) {
	if (level == SOL_SOCKET &&
		(optname == SO_SNDBUF || optname == SO_RCVBUF ||
		optname == SO_BROADCAST)) {

	    prefault();
	    vista_set_range(dc_seg_ptr, &(explicit_ptr->fds[s].u.sock),
			    sizeof(explicit_ptr->fds[s].u.sock),
			    explicit_ptr->dc_tid);
	    if (optname == SO_SNDBUF) {
		explicit_ptr->fds[s].u.sock.sndbuf = * (int *)optval;
	    } else if (optname == SO_RCVBUF) {
		explicit_ptr->fds[s].u.sock.rcvbuf = * (int *)optval;
	    } else if (optname == SO_BROADCAST) {
		explicit_ptr->fds[s].u.sock.broadcast = * (int *)optval;
	    }
	} else {
	    fprintf(stderr, "warning: setsockopt won't be restored during recovery: s=%d, level=%d, optname=%d, *optval=%d\n",
			    s, level, optname, * (int *)optval);
	}
    }
    return(syscall(SYS_setsockopt, s, level, optname, optval, optlen));
}

int getsockopt(int s, int level, int optname, void *optval, int *optlen)
{
    int status;

    dc_lock();

    if (optlen != NULL) {
	dc_touch((char *) optval, *optlen);
	dc_touch((char *) optlen, sizeof(int));
    }
    status = syscall(SYS_getsockopt, s, level, optname, optval, optlen);

    dc_unlock();

    return(status);
}

int bind(int s, const struct sockaddr *name, int namelen)
{
    int status;
    int length;
    int port;

    port = ((struct sockaddr_in *) name)->sin_port;

    status = syscall(SYS_bind, s, name, namelen);

    if (explicit_ptr->fds[s].status != FD_SOCKET) {
	fprintf(stderr, "warning: binding non-DC socket fd %d\n", s);
	return(status);
    }

    if (status == 0) {
	prefault();
	vista_set_range(dc_seg_ptr, &(explicit_ptr->fds[s].u.sock),
			sizeof(explicit_ptr->fds[s].u.sock),
			explicit_ptr->dc_tid);
	explicit_ptr->fds[s].u.sock.bound = 1;

	/*
	 * Use getsockname to get bound address in case bind called with port 0
	 * (in which case the system picks an address).
	 */
	length = sizeof(explicit_ptr->fds[s].u.sock.addr);
	if (getsockname(s,
	    (struct sockaddr *) &explicit_ptr->fds[s].u.sock.addr, &length)) {
	    perror("getsockname");
	    exit(1);
	}
    }

    /*
     * Bind is non-deterministic if the user asked the system to choose a port.
     */
    if (port == 0) {
	seen_nd("bind");
    }

    return(status);
}

/*
 * DC implements connect by remembering the destination address for each
 * connected socket, then filling in that address for sends (sendto with NULL
 * destination addresses).  The alternative of using the connect system call
 * doesn't work, because recvmsg of a connected socket can return ECONNREFUSED
 * (which is very strange for a datagram socket).
 */
int connect(int s, const struct sockaddr *name, int namelen)
{
    if (explicit_ptr->fds[s].status != FD_SOCKET) {
	fprintf(stderr, "warning: connecting non-DC socket fd %d\n", s);
	return(syscall(SYS_connect, s, name, namelen));
    }

    prefault();
    vista_set_range(dc_seg_ptr, &(explicit_ptr->fds[s].u.sock),
		    sizeof(explicit_ptr->fds[s].u.sock),
		    explicit_ptr->dc_tid);
    if (name == NULL) {
	/* NULL name pointer tells system to disconnect the socket */
	explicit_ptr->fds[s].u.sock.connected = 0;
    } else {
	explicit_ptr->fds[s].u.sock.connected = 1;
	bcopy(name, &explicit_ptr->fds[s].u.sock.connaddr,
		sizeof(explicit_ptr->fds[s].u.sock.connaddr));
    }

    return(0);
}

/*
 * Check if a socket is bound yet.  If not, system assigns an address.
 * Use bind (not syscall(SYS_bind)) so DC can record the bound address.
 * This should be called before a send/sendto/sendmsg/connect, since that's the
 * time the system picks an address if not yet bound.
 */
static int check_bound(int s)
{
    struct sockaddr_in addr;

    if (explicit_ptr->fds[s].status == FD_SOCKET &&
	    !explicit_ptr->fds[s].u.sock.bound) {

	bzero(&addr, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = htonl(INADDR_ANY);
	addr.sin_port = htons(0);
	if (bind(s, (struct sockaddr*) &addr, sizeof(addr))) {
	    perror("bind (in check_bound)");
	    exit(1);
	}
    }
}

/*
 * Copy length bytes of buffer into an iov struct (but not more than the
 * iov struct can hold).  Return the number of bytes copied.
 */
int copy_to_iov(char *buf, int length, struct iovec *iov, int iovcnt)
{
    int status = 0; /* # of bytes copied so far */
    int i;
    int size;

    for (i = 0; status < length && i < iovcnt; i++) {
	size = length - status;
	if (size > iov[i].iov_len) {
	    size = iov[i].iov_len;
	}
	memcpy(iov[i].iov_base, buf + status, size);
	status += size;
    }
    return(status);
}

/*
 * Copy length bytes from an iov struct to buffer.  Return the number of bytes
 * copied.
 */
int copy_from_iov(char *buf, int length, struct iovec *iov, int iovcnt)
{
    int status = 0; /* # of bytes copied so far */
    int i;
    int size;

    for (i = 0; status < length && i < iovcnt; i++) {
	size = length - status;
	if (size > iov[i].iov_len) {
	    size = iov[i].iov_len;
	}
	memcpy(buf + status, iov[i].iov_base, size);
	status += size;
    }
    return(status);
}

/*
 * Check the redo log to see if there's a record to replay.  Advance the
 * redo_tail_ptr over invalid records.
 *
 * Return 1 if there's a redo record to replay, 0 if no record to replay.
 * If it finds a redo record to replay, check_redo sets redo_tail_ptr to
 * point to that record.  Otherwise, it leaves redo_tail_ptr pointing at the
 * last record on the log.
 */
int check_redo()
{
    while (explicit_ptr->redo_tail_ptr->next_ptr != NULL &&
	    explicit_ptr->redo_tail_ptr->next_ptr->type == REDO_INVALID) {
	explicit_ptr->redo_tail_ptr = explicit_ptr->redo_tail_ptr->next_ptr;
    }

    if (explicit_ptr->redo_tail_ptr->next_ptr != NULL) {
	/*
	 * Found a redo record to replay.
	 */
	explicit_ptr->redo_tail_ptr = explicit_ptr->redo_tail_ptr->next_ptr;
	return(1);
    } else {
	return(0);
    }
}

/*
 * Trap select so that the extra SIG2PC signals are automatically restarted.
 *
 * RMessages also traps select, so we need to rename DC's version dc_select
 * (called by RMessages) when DC and RMessages are used together.
 */
#ifdef RMESSAGES
int dc_select(int nfds, fd_set *readfds, fd_set *writefds, fd_set *exceptfds,
    struct timeval *timeout)
#else
int select(int nfds, fd_set *readfds, fd_set *writefds, fd_set *exceptfds,
    struct timeval *timeout)
#endif
{
    int status;
    int repeat;
    int fd;
    selbuf *sel_ptr;
    fd_set prior_fds;
    int prior_fd_flag = 0;
    struct timeval t;
    struct timeval *t_ptr = timeout;

    /* ??? hack to get around bug in rmessages, which calls select with
	nfds=4096 */
    if (nfds > FD_SETSIZE) {
	nfds = FD_SETSIZE;
    }

    /*
     * Check each select buffer to see which of the readfds should be
     * set by DC.
     */
    FD_ZERO(&prior_fds);

    for (fd = 0; readfds != NULL & fd < nfds; fd++) {
	if (FD_ISSET(fd, readfds) &&
		(explicit_ptr->fds[fd].status == FD_SOCKET ||
		explicit_ptr->fds[fd].status == FD_SPECIAL) &&
		explicit_ptr->fds[fd].select_buf != NULL) {
	    /*
	     * Already something pending on this port, so don't bother
	     * selecting it.
	     */
	    FD_CLR(fd, readfds);
	    FD_SET(fd, &prior_fds);

	    prior_fd_flag = 1;

	    /*
	     * Don't block, since I already know there's something ready.
	     */
	    t_ptr = &t;
	    t_ptr->tv_sec = t_ptr->tv_usec = 0;
	}
    }

    do {
	dc_lock();

	dc_touch((char *) readfds, sizeof(*readfds));
	dc_touch((char *) writefds, sizeof(*writefds));
	dc_touch((char *) exceptfds, sizeof(*exceptfds));

	/*
	 * If there was a signal, treat it as though it came in during
	 * the system call.  This minimizes the chance that we'll start the
	 * select call with a pending deferred signal.  If by chance a
	 * signal comes in between the if and the syscall, then it's just like
	 * a dropped message.
	 *
	 * Give the select a chance if the deferred signal is SIGIO, because
	 * select in a sigio handler (as in treadmarks) may KNOW there is
	 * something on the port and call select (and not expect it to get a
	 * SIGIO)
	 */
	if (defer_any && !defer[SIGIO]) {
	    status = -1;
	    errno = EINTR;
	    fprintf(stderr, "bypassing select call\n"); /* ??? debugging
								output */
	} else {
	    status = syscall(SYS_select, nfds, readfds, writefds, exceptfds,
				t_ptr);
	}

	/* ??? debugging output */
	if (status < 0) {
	    fprintf(stderr, "select failed, errno = %d\n", errno);
	    for (fd = 1; fd < NSIG; fd++) {
		if (defer[fd]) {
		    fprintf(stderr, "signal %d was set\n", fd);
		}
	    }
	}

	/*
	 * Restart the call if I got a SIG2PC signal.  Select doesn't restart
	 * from other signals (even those registered SA_RESTART).
	 */
	if (status < 0 && errno == EINTR && defer[SIG2PC]) {
	    fprintf(stderr, "repeating select due to a SIG2PC signal\n");
	    repeat = 1;
	} else {
	    repeat = 0;
	}

	dc_unlock();

    } while (repeat);

    /*
     * Save the first pending message on each port.  Otherwise a checkpoint
     * doesn't capture the entire (visible) state of the process.  E.g.
     *   a. select informs the user that a port is ready
     *   b. checkpoint
     *   c. program crashes and restarts.  The next recv on the ready port
     *      will block (even though the program was told that the port had
     *      data waiting on it).
     * Only need to save pending message on sockets and devices (select is
     * deterministic for files).
     * This save is done TRANS (vista_malloc is done TRANS, so don't have
     * to set_range the buffer).
     */

    if (status >= 0 && readfds != NULL) {
	for (fd = 0; fd < nfds; fd++) {
	    if (FD_ISSET(fd, readfds) &&
		    explicit_ptr->fds[fd].select_buf == NULL &&
		    (explicit_ptr->fds[fd].status == FD_SOCKET ||
		    explicit_ptr->fds[fd].status == FD_SPECIAL) ) {

		if (explicit_ptr->fds[fd].status == FD_SOCKET) {
		    sel_ptr = vista_malloc(explicit_seg_ptr, sizeof(selbuf),
					    explicit_ptr->dc_tid);
		    sel_ptr->buf = vista_malloc(explicit_seg_ptr, MAX_MESSAGE,
						explicit_ptr->dc_tid);
		    sel_ptr->fromlen = sizeof(sel_ptr->from);
		    sel_ptr->length = syscall(SYS_recvfrom, fd, sel_ptr->buf,
						MAX_MESSAGE, 0, &sel_ptr->from,
						&sel_ptr->fromlen);
		} else if (explicit_ptr->fds[fd].status == FD_SPECIAL) {
		    sel_ptr = vista_malloc(explicit_seg_ptr, sizeof(selbuf),
				explicit_ptr->dc_tid);
		    sel_ptr->buf = vista_malloc(explicit_seg_ptr, SPECIAL_BLOCK,
						explicit_ptr->dc_tid);
		    sel_ptr->length = syscall(SYS_read, fd, sel_ptr->buf,
						SPECIAL_BLOCK);
		}
		vista_set_range(dc_seg_ptr, &(explicit_ptr->fds[fd].select_buf),
				sizeof(explicit_ptr->fds[fd].select_buf),
				explicit_ptr->dc_tid);
		explicit_ptr->fds[fd].select_buf = sel_ptr;

	    } else {
		/*
		 * Account for those fds that already had saved select buffers.
		 */
		if (FD_ISSET(fd, &prior_fds)) {
		    FD_SET(fd, readfds);
		    status++;
		}
	    }
	}
    }

    /*
     * Select is non-deterministic if there are any read fds.
     */
    if (readfds != NULL) {
	seen_nd("select");
    }

    return(status);
}

ssize_t sendmsg(int s, const struct msghdr *msg, int flags)
{
    struct msghdr msg1;
    int status;
    int i;

#ifdef TWO_PHASE_COMMIT
    /*
     * Set the sent bit.  This needs to be NOTRANS, because recovery needs
     * to know that a message was set.
     */
    explicit_ptr->sent_message = 1;
#endif

    if (explicit_ptr->fds[s].status != FD_SOCKET) {
	if (s >= USER_FD) {
	    fprintf(stderr, "warning: sendmsg non-DC socket fd %d\n", s);
	}
	return(syscall(SYS_sendmsg, s, msg, flags));
    }
    check_bound(s);

    msg1 = *msg;

    if (msg1.msg_name == NULL && explicit_ptr->fds[s].u.sock.connected) {
	msg1.msg_name = (caddr_t) &explicit_ptr->fds[s].u.sock.connaddr;
	msg1.msg_namelen = sizeof(explicit_ptr->fds[s].u.sock.connaddr);
    }

    /*
     * Send is non-abortable (unless using TWO_PHASE_COMMIT).  So checkpoint
     * if there's been a non-deterministic since the last checkpoint.
     * Only checkpoint if we're tracking non-abortables.
     *
     * Problems could occur if a signal occurs between dc_checkpoint
     * and the send.  The signal (a non-deterministic event) could
     * theoretically change the data that gets sent out and make the
     * execution during recovery not match the execution before the crash.
     * This doesn't seem worth fixing for now. ???
     */
/* ??? special case for sends to X server for xpilot.  Sends to the X server
    are considered visible events */
#ifdef TWO_PHASE_COMMIT
#ifdef DC_DISK
    if (s == 13)
#else
    if (s == 10)
#endif /* DC_DISK */
    {
	seen_na("sendmsg");
    }
#endif /* TWO_PHASE_COMMIT */



#if !defined(TWO_PHASE_COMMIT)
    seen_na("sendmsg");
#endif

    status = syscall(SYS_sendmsg, s, &msg1, flags);
    if (status < 0) {
	fprintf(stderr, "sendmsg failed\n");
	perror("sendmsg");
    }

    return(status);
}

/*
 * Intercept sendto (and send, since send calls sendto) by transforming it
 * into sendmsg.
 */
ssize_t sendto(int s, const void *buf, size_t len, int flags,
		  const struct sockaddr *to, int tolen)
{
    struct msghdr msg;
    struct iovec iov;

    msg.msg_name = (caddr_t) to;
    msg.msg_namelen = tolen;
    msg.msg_iov = &iov;
    msg.msg_iovlen = 1;
    msg.msg_control = NULL;
    msg.msg_controllen = 0;
    msg.msg_flags = 0; /* result parameter */

    iov.iov_base = (char *) buf;
    iov.iov_len = len;

    return(sendmsg(s, &msg, flags));
}

/*
 * recv/recvfrom/recvmsg are intercepted to ensure that the buffers and return
 * address are pre-faulted, and to implement SA_RESTART semantics.
 */
ssize_t recvmsg(int s, struct msghdr *msg, int flags)
{
    int i;
    int status;
    int repeat;
    selbuf *sel_ptr = NULL;
    redo_node *redo_ptr;
    int redo_flag = 0;

    /*
     * Check for a previously saved message from a port that user called
     * select() on.  Do this with dc_lock held so a signal can't come in
     * and cause a concurrent access to the select buffer.
     */
    dc_lock();

    sel_ptr = explicit_ptr->fds[s].select_buf;
    if (sel_ptr != NULL && explicit_ptr->fds[s].status == FD_SOCKET) {
	/*
	 * Copy select buffer into msg.
	 */
	if (msg->msg_name != NULL) {
	    memcpy(msg->msg_name, &sel_ptr->from, sel_ptr->fromlen);
	    msg->msg_namelen = sel_ptr->fromlen;
	}

	status = copy_to_iov(sel_ptr->buf, sel_ptr->length, msg->msg_iov,
				msg->msg_iovlen);

	/*
	 * Mark the select buffer as empty (do this during dc_lock region
	 * to ensure that nobody else uses this buffer).  Partially received
	 * buffers are discarded (as usual with recv).  Do this TRANS so
	 * the select buffer reappears if a crash happens before the next
	 * checkpoint.
	 *
	 * Don't delete the select buffer if MSG_PEEK.
	 */
	if (! (flags & MSG_PEEK) ) {
	    vista_set_range(dc_seg_ptr, &(explicit_ptr->fds[s].select_buf),
			    sizeof(explicit_ptr->fds[s].select_buf),
			    explicit_ptr->dc_tid);
	    explicit_ptr->fds[s].select_buf = NULL;

	    dc_unlock();

	    vista_free(explicit_seg_ptr, sel_ptr->buf, explicit_ptr->dc_tid);
	    vista_free(explicit_seg_ptr, sel_ptr, explicit_ptr->dc_tid);
	} else {
	    dc_unlock();
	}

	/*
	 * Receives from select buffers are deterministic, so they don't
	 * have to go through the logging code below, or call seen_nd.
	 */
	return(status);
    }

    dc_unlock();

#ifdef REDO_LOG
    /*
     * Create the redo record.
     */
    redo_ptr = vista_malloc(explicit_seg_ptr, sizeof(redo_node),
			    VISTA_NO_TRANS);
    redo_ptr->type = REDO_INVALID;
    redo_ptr->next_ptr = NULL;
#endif

    do {
	dc_lock();

	/*
	 * Check to see if we're replaying the redo log.  This check_redo must
	 * be atomic with the receive call.  Otherwise a checkpoint might
	 * happen before the receive.  In this case, there will be (and should
	 * be) a redo record, since the receive occured after the checkpoint.
	 * But the checkpoint will recover to a point AFTER check_redo, so
	 * it won't check for and use the redo record.
	 */
	if (check_redo()) {
	    fprintf(stderr, "recvmsg replaying logged message\n");

	    redo_ptr = explicit_ptr->redo_tail_ptr;

	    /*
	     * Check parameters to make sure we're still in sync with pre-crash.
	     */
	    if (redo_ptr->type != REDO_RECEIVE || redo_ptr->fd != s ||
		    redo_ptr->buf != (char *) msg) {
		fprintf(stderr, "error: recvmsg replay not in sync\n");
		fprintf(stderr, "redo_ptr->type = %d (%d), redo_ptr->fd = %d (%d), redo_ptr->buf = %x (%x)\n", redo_ptr->type, REDO_RECEIVE, redo_ptr->fd, s, redo_ptr->buf, (char *) msg);
		exit(1);
	    }
	    if (redo_ptr->return_status < 0) {
		errno = redo_ptr->return_errno;
	    } else {
		if (msg->msg_name != NULL) {
		    memcpy(msg->msg_name, &redo_ptr->select_buf_ptr[0]->from,
			redo_ptr->select_buf_ptr[0]->fromlen);
		}
		msg->msg_namelen = redo_ptr->select_buf_ptr[0]->fromlen;

		/*
		 * Copy received data from redo log record.
		 */
		copy_to_iov(redo_ptr->select_buf_ptr[0]->buf,
			    redo_ptr->select_buf_ptr[0]->length,
			    msg->msg_iov, msg->msg_iovlen);
	    }

	    dc_unlock();

	    return(redo_ptr->return_status);
	}

#ifdef REDO_LOG
	/*
	 * Add the redo record to the redo log (but don't make
	 * it valid yet).  This must be atomic with the receive call;
	 * otherwise a signal/checkpoint could occur after adding to the
	 * redo log (which would truncate the redo log) but before the receive.
	 * This would cause the subsequent receive to not be logged (which is
	 * wrong).  The redo record won't be valid until the receive is
	 * finished, so this record won't get used yet even though it's
	 * connected on the redo log.
	 *
	 * Adding to the redo log is done NO_TRANS, since I want the new redo
	 * log records to live if a crash happens before the next checkpoint.
	 *
	 * Don't add to the redo log if there's been a non-deterministic event
	 * since the last checkpoint, since I'm not guaranteed to go down this
	 * path.
	 */
	redo_flag = 0;
	if (!explicit_ptr->seen_nd && explicit_ptr->fds[s].status == FD_SOCKET) {
	    /*
	     * Attach it to the redo log (though it's still invalid).  Don't
	     * advance the redo_tail yet, because I might go through this loop
	     * again and re-attach the record.
	     */
	    explicit_ptr->redo_tail_ptr->next_ptr = redo_ptr;
	    redo_flag = 1;
	}
#endif

	for (i = 0; i < msg->msg_iovlen; i++) {
	    dc_touch(msg->msg_iov[i].iov_base, msg->msg_iov[i].iov_len);
	}
	dc_touch(msg->msg_name, msg->msg_namelen);
	dc_touch((char *) msg, sizeof(*msg));
	dc_touch((char *) msg->msg_control, msg->msg_controllen);

	/*
	 * If there was a signal, treat it as though it came in during
	 * the system call.  This minimizes the chance that we'll start the
	 * recvmsg call with a pending deferred signal.  If by chance a
	 * signal comes in between the if and the syscall, then it's just like
	 * a dropped message.
	 */
	if (defer_any) {
	    status = -1;
	    errno = EINTR;
	} else {
	    status = syscall(SYS_recvmsg, s, msg, flags);
	}

	/*
	 * Implement SA_RESTART by restarting the call.  The system's restart
	 * mechanism would just restart the syscall, which wouldn't include the
	 * dc_touch, and wouldn't allow the vista_unlock call (which actually
	 * processes the deferred signal).  I need to process the signal,
	 * then restart the call.  If any of the deferred signals are not marked
	 * SA_RESTART, I assume that was the one that interrupted the system,
	 * so I won't restart the call (it may well have been that signal, and
	 * there's no possibility of guaranteed ordering between multiple
	 * signals within a vista_lock region, since signals get deferred
	 * (not processed)).
	 */
	if (status < 0 && errno == EINTR && !sig_norestart()) {
	    repeat = 1;
	} else {
	    repeat = 0;
	    /*
	     * Now that the receive is done (and won't be re-done), advance the
	     * redo_tail.  If a signal/checkpoint happens at the next
	     * dc_unlock, it will truncate the redo log.
	     */
	    if (redo_flag) {
		explicit_ptr->redo_tail_ptr = redo_ptr;
	    }
	}

	dc_unlock();

    } while (repeat);

    if (status < 0 && errno == EFAULT) {
	perror("recvmsg");
	exit(1);
    }

    /*
     * Log messages.  If messages are not logged, then receiving a message is
     * non-deterministic.
     * If a checkpoint happened after the receive, this redo record will have
     * been disconnected during the checkpoint (so it won't be used).  The
     * checkpoint won't free an invalid entry, so the memory is still there
     * (so this won't core dump).
     */
#ifdef REDO_LOG
    if (redo_flag) {
	if (redo_ptr == NULL) {
	    fprintf(stderr, "error: redo_ptr = NULL\n");
	    exit(1);
	}
	redo_ptr->fd = s;
	redo_ptr->buf = (char *) msg;
	redo_ptr->return_status = status;
	if (status < 0) {
	    redo_ptr->return_errno = errno;
	    redo_ptr->select_buf_ptr[0] = NULL;
	} else {
	    sel_ptr = redo_ptr->select_buf_ptr[0] =
		vista_malloc(explicit_seg_ptr, sizeof(selbuf), VISTA_NO_TRANS);
	    sel_ptr->buf = vista_malloc(explicit_seg_ptr, status,
					VISTA_NO_TRANS);

	    if (sel_ptr->buf == NULL) {
		fprintf(stderr, "error in recvmsg: vista_malloc failed\n");
		exit(1);
	    }

	    if (msg->msg_name != NULL) {
		sel_ptr->from = *((struct sockaddr_in *) msg->msg_name);
	    }
	    sel_ptr->fromlen = msg->msg_namelen;
	    sel_ptr->length = status;

	    /*
	     * Copy received data to redo log record.
	     */
	    copy_from_iov(sel_ptr->buf, status, msg->msg_iov, msg->msg_iovlen);

	    /*
	     * Free this redo record if there was a checkpoint anytime
	     * after the receive.  This isn't necessary, but it decreases
	     * memory leakage.
	     */
	    if (explicit_ptr->redo_tail_ptr == &explicit_ptr->redo_head) {
		vista_free(explicit_seg_ptr, sel_ptr->buf, VISTA_NO_TRANS);
		vista_free(explicit_seg_ptr, sel_ptr, VISTA_NO_TRANS);
		vista_free(explicit_seg_ptr, redo_ptr, VISTA_NO_TRANS);
	    }
	}

	/*
	 * This needs to be done AFTER the record is completely initialized.
	 */
	redo_ptr->type = REDO_RECEIVE;

#ifndef TRACK_NA
	/*
	 * Flush the redo log to disk (only has effect if using DC_DISK).
	 * This is done if we're not tracking non-abortables, since then we
	 * must assume that the next event is non-abortable.
	 */
	redo_flush();
#endif
    } else {
	/*
	 * The redo record wasn't attached to the redo log, so free it.
	 */
	vista_free(explicit_seg_ptr, redo_ptr, VISTA_NO_TRANS);
    }

#else

    seen_nd("recvmsg");

#endif /* REDO_LOG */

    return(status);
}

/*
 * Intercept recvfrom (and recv, since recv calls recvfrom) by transforming it
 * into recvmsg.
 */
ssize_t recvfrom(int s, void *buf, size_t len, int flags, struct sockaddr *from,
      int *fromlen)
{
    struct msghdr msg;
    struct iovec iov;

    msg.msg_name = (caddr_t) from;
    if (fromlen == NULL) {
	msg.msg_namelen = 0;
    } else {
	msg.msg_namelen = *fromlen;
    }
    msg.msg_iov = &iov;
    msg.msg_iovlen = 1;
    msg.msg_control = NULL;
    msg.msg_controllen = 0;
    msg.msg_flags = 0; /* result parameter */

    iov.iov_base = buf;
    iov.iov_len = len;

    return(recvmsg(s, &msg, flags));
}

/*
 * Check if there was a deferred signal that was not marked SA_RESTART.
 */
static int sig_norestart()
{
    int i;

    for (i=1; i<NSIG; i++) {
	if (defer[i] && !(siglist[i].sa_flags & SA_RESTART) ) {
	    return(1);
	}
    }
    return(0);
}

/*
 * Trap fcntl to remember if a socket was set to generate SIGIO.
 * <fcntl.h> has this as a varargs--not sure why.
 */
int fcntl(int fd, int cmd, ...)
{
    va_list ap;
    int arg;

    va_start(ap, cmd);
    arg = va_arg(ap, int);
    va_end(ap);

    if (explicit_ptr->fds[fd].status == FD_SOCKET && cmd == F_SETFL &&
	arg == O_ASYNC) {
	fprintf(stderr, "marking fd %d to generate SIGIO to current process\n",
	    fd);
	prefault();
	vista_set_range(dc_seg_ptr, &(explicit_ptr->fds[fd].u.sock.sigio),
			sizeof(explicit_ptr->fds[fd].u.sock.sigio),
			explicit_ptr->dc_tid);
	explicit_ptr->fds[fd].u.sock.sigio = 1;
    } else if (explicit_ptr->fds[fd].status == FD_SOCKET && cmd == F_SETFL &&
	arg == O_NONBLOCK) {
	fprintf(stderr, "marking fd %d to be non-blocking\n", fd);
	prefault();
	vista_set_range(dc_seg_ptr, &(explicit_ptr->fds[fd].u.sock.nonblock),
			sizeof(explicit_ptr->fds[fd].u.sock.nonblock),
			explicit_ptr->dc_tid);
	explicit_ptr->fds[fd].u.sock.nonblock = 1;
    } else if (cmd == F_GETFD || cmd == F_SETFD || cmd == F_GETFL ||
	    cmd == F_GETOWN) {
	/* these don't need to be recoverable */
    } else if (cmd == F_SETOWN) {
	/*
	 * Need to trap this to translate the old pid into the new pid.
	 */
	arg = get_current_pid(arg);
    } else {
	fprintf(stderr, "warning: fcntl(fd=%d, cmd=%d, arg=%d) not recoverable\n",
	    fd, cmd, arg);
    }
    return(syscall(SYS_fcntl, fd, cmd, arg));
}

/*
 * Implement unlink so that deleted-but-open files aren't deleted when the
 * process exits.  This way the process can recover and still have the files
 * present.  Files marked as unlinked are deleted when they're closed.
 */
int unlink(const char *path)
{
    int i;

    for (i=0; explicit_ptr && i<OPEN_MAX; i++) {
	if (explicit_ptr->fds[i].status == FD_FILE &&
		!strcmp(path, explicit_ptr->fds[i].u.file.path)) {
	    prefault();
	    vista_set_range(dc_seg_ptr, &(explicit_ptr->fds[i].u.file.unlink),
			sizeof(explicit_ptr->fds[i].u.file.unlink),
			explicit_ptr->dc_tid);
	    explicit_ptr->fds[i].u.file.unlink = 1;
	    return(0);
	}
    }

    /*
     * If this file wasn't open, unlink it right away.
     */
    return(syscall(SYS_unlink, path));
}

/*
 * Make an lseek system call.  Code taken from /usr/src/lib/libc/sys/lseek.c
 * This code is different than a standard syscall because of the 64-bit off_t.
 */
static off_t sys_lseek(int d, off_t offset, int whence)
{
    return(__syscall((quad_t)SYS_lseek, d, 0, offset, whence));
}

/*
 * Intercept lseek to maintain current file position.
 */
off_t lseek(int d, off_t offset, int whence)
{
    off_t status;

    status = sys_lseek(d, offset, whence);

    /*
     * Update current file position if it's changed.
     * The check for d == STDOUT_FILENO or d > USER_FD is used because this
     * lseek is called before the explicit segment is initialized (so we can't
     * just depend on explicit_ptr->fds[d].status).
     */
    if ( (d == STDOUT_FILENO || d >= USER_FD) && status >= 0 &&
	    explicit_ptr->fds[d].status == FD_FILE &&
	    status != explicit_ptr->fds[d].u.file.position) {
	    
	prefault();
	vista_set_range(dc_seg_ptr, &(explicit_ptr->fds[d].u.file.position),
		    sizeof(explicit_ptr->fds[d].u.file.position),
		    explicit_ptr->dc_tid);
	explicit_ptr->fds[d].u.file.position = status;
    }
    return(status);
}

ssize_t readv(int d, const struct iovec *iov, int iovcnt)
{
    int i;
    int status;
    int repeat;
    struct msghdr msg;

    selbuf *sel_ptr = NULL;
    char *old_buf;
    int size;
    redo_node *redo_ptr;
    int redo_flag = 0;
    int keyboard = 0;
    int nondeterm;

    /*
     * We use bit 31 in iov[0].iov_len to indicate that this read is a
     * simulated keyboard.  This is a hack so that simulated keyboard input
     * (which actually comes from a file) gets treated as a non-deterministic
     * event or is logged.
     */
    if (iovcnt > 0 && (iov[0].iov_len & 0x80000000)) {
	keyboard = 1;
	/* this mess is a typecast to get around the const iov */
	* (size_t *) (& (iov[0].iov_len)) &= 0x7fffffff;
    }

    /*
     * Is this readv call non-deterministic?
     */
    nondeterm = (keyboard || explicit_ptr->fds[d].status== FD_SPECIAL);

    /*
     * Don't dc_touch readv initiated by vista routines, since we'd need to get
     * the vista lock and it's already held by the vista routine.  The vista
     * routines guarantee no prot faults anyway.
     */
    if (d < USER_FD && d != 0) {
	return(syscall(SYS_readv, d, iov, iovcnt));
    }

    /*
     * Re-direct readv on sockets to recvmsg to take care of connected sockets
     */
    if (explicit_ptr->fds[d].status == FD_SOCKET) {
	msg.msg_name = NULL;
	msg.msg_namelen = 0;
	msg.msg_iov = (struct iovec *) iov; /* typecast to drop const warning */
	msg.msg_control = NULL;
	msg.msg_controllen = 0;
	return(recvmsg(d, &msg, 0));
    }

    /*
     * Check for a previously saved message from a port that user called
     * select() on.  Do this with dc_lock held so a signal can't come in
     * and cause a concurrent access to the select buffer.
     */
    dc_lock();

    sel_ptr = explicit_ptr->fds[d].select_buf;
    if (sel_ptr != NULL && explicit_ptr->fds[d].status == FD_SPECIAL) {
	/*
	 * Copy select buffer into iov
	 */
	status = copy_to_iov(sel_ptr->buf, sel_ptr->length,
			    (struct iovec *) iov, iovcnt);

	if (sel_ptr->length == status) {
	    /*
	     * Used up all the bytes, so discard the select buffer.
	     */
	    vista_set_range(dc_seg_ptr, &(explicit_ptr->fds[d].select_buf),
			    sizeof(explicit_ptr->fds[d].select_buf),
			    explicit_ptr->dc_tid);
	    explicit_ptr->fds[d].select_buf = NULL;

	    dc_unlock();

	    vista_free(explicit_seg_ptr, sel_ptr->buf, explicit_ptr->dc_tid);
	    vista_free(explicit_seg_ptr, sel_ptr, explicit_ptr->dc_tid);

	} else {
	    /*
	     * Copy the remaining bytes to a new data buffer.
	     */
	    vista_set_range(dc_seg_ptr, sel_ptr, sizeof(*sel_ptr),
			    explicit_ptr->dc_tid);

	    size = sel_ptr->length - status;
	    old_buf = sel_ptr->buf;
	    sel_ptr->buf = vista_malloc(explicit_seg_ptr, size,
					explicit_ptr->dc_tid);

	    memcpy(sel_ptr->buf, old_buf + status, size);

	    sel_ptr->length = size;

	    dc_unlock();

	    vista_free(explicit_seg_ptr, old_buf, explicit_ptr->dc_tid);
	}

	/*
	 * Reads from select buffers are deterministic, so they don't
	 * have to go through the logging code below, or call seen_nd.
	 */
	return(status);
    }

    dc_unlock();

#ifdef REDO_LOG
    /*
     * Create the redo record.
     */
    if (nondeterm) {
	redo_ptr = vista_malloc(explicit_seg_ptr, sizeof(redo_node),
				VISTA_NO_TRANS);
	redo_ptr->type = REDO_INVALID;
	redo_ptr->next_ptr = NULL;
    }
#endif

    do {
	dc_lock();

	/*
	 * Check to see if we're replaying the redo log.  This check_redo must
	 * be atomic with the receive call.  Otherwise a checkpoint might
	 * happen before the read.  In this case, there will be (and should
	 * be) a redo record, since the receive occured after the checkpoint.
	 * But the checkpoint will recover to a point AFTER check_redo, so
	 * it won't check for and use the redo record.
	 */
	if (nondeterm && check_redo()) {
	    /* fprintf(stderr, "readv replaying logged input\n"); */

	    redo_ptr = explicit_ptr->redo_tail_ptr;

	    /*
	     * Check parameters to make sure we're still in sync with pre-crash.
	     */
	    if (redo_ptr->type != REDO_READ || redo_ptr->fd != d ||
		    redo_ptr->buf != (char *) iov) {
		fprintf(stderr, "error: readv replay not in sync\n");
		fprintf(stderr, "redo_ptr->type = %d (%d), redo_ptr->fd = %d (%d), redo_ptr->buf = %x (%x)\n", redo_ptr->type, REDO_READ, redo_ptr->fd, d, redo_ptr->buf, (char *) iov);
		exit(1);
	    }
	    if (redo_ptr->return_status <= 0) {
		errno = redo_ptr->return_errno;
	    } else {
		/*
		 * Copy input data from redo log record.
		 */
		copy_to_iov(redo_ptr->select_buf_ptr[0]->buf,
			    redo_ptr->select_buf_ptr[0]->length,
			    (struct iovec *) iov, iovcnt);
	    }

	    dc_unlock();

	    /*
	     * Update position if readv succeeded.
	     */
	    if (explicit_ptr->fds[d].status == FD_FILE &&
		    redo_ptr->return_status > 0) {
		prefault();
		vista_set_range(dc_seg_ptr,
				&(explicit_ptr->fds[d].u.file.position),
				sizeof(explicit_ptr->fds[d].u.file.position),
				explicit_ptr->dc_tid);
		explicit_ptr->fds[d].u.file.position += redo_ptr->return_status;

		/*
		 * Re-position file as it was after the replayed read (this
		 * will only be done for files that are actually simulated
		 * devices).
		 */
		sys_lseek(d, (off_t) redo_ptr->return_status, SEEK_CUR);
	    }

	    return(redo_ptr->return_status);
	}

#ifdef REDO_LOG
	/*
	 * Add the redo record to the redo log (but don't make
	 * it valid yet).  This must be atomic with the read call;
	 * otherwise a signal/checkpoint could occur after adding to the
	 * redo log (which would truncate the redo log) but before the read.
	 * This would cause the subsequent read to not be logged (which is
	 * wrong).  The redo record won't be valid until the read is
	 * finished, so this record won't get used yet even though it's
	 * connected on the redo log.
	 *
	 * Adding to the redo log is done NO_TRANS, since I want the new redo
	 * log records to live if a crash happens before the next checkpoint.
	 *
	 * Don't add to the redo log if there's been a non-deterministic event
	 * since the last checkpoint, since I'm not guaranteed to go down this
	 * path.
	 */
	redo_flag = 0;
	if (nondeterm && !explicit_ptr->seen_nd) {
	    /*
	     * Attach it to the redo log (though it's still invalid).  Don't
	     * advance the redo_tail yet, because I might go through this loop
	     * again and re-attach the record.
	     */
	    explicit_ptr->redo_tail_ptr->next_ptr = redo_ptr;
	    redo_flag = 1;
	}
#endif

	for (i=0; i<iovcnt; i++) {
	    dc_touch(iov[i].iov_base, iov[i].iov_len);
	}

	/*
	 * If there was a signal, treat it as though it came in during
	 * the system call.
	 */
	if (defer_any) {
	    status = -1;
	    errno = EINTR;
	} else {
	    status = syscall(SYS_readv, d, iov, iovcnt);
	}

	/* Implement SA_RESTART by restarting the call */
	if (status < 0 && errno == EINTR && !sig_norestart()) {
	    repeat = 1;
	} else {
	    repeat = 0;
	    /*
	     * Now that the read is done (and won't be re-done), advance the
	     * redo_tail.  If a signal/checkpoint happens at the next
	     * dc_unlock, it will truncate the redo log.
	     */
	    if (redo_flag) {
		explicit_ptr->redo_tail_ptr = redo_ptr;
	    }
	}

	dc_unlock();

    } while (repeat);

    /*
     * Update position if readv succeeded.
     */
    if (explicit_ptr->fds[d].status == FD_FILE && status >= 0) {
	prefault();
	vista_set_range(dc_seg_ptr, &(explicit_ptr->fds[d].u.file.position),
		    sizeof(explicit_ptr->fds[d].u.file.position),
		    explicit_ptr->dc_tid);
	explicit_ptr->fds[d].u.file.position += status;
    }

    /*
     * Log reads of devices and simulated keyboards.  If reads are not logged,
     * then reading input is non-deterministic.
     * If a checkpoint happened after the read, this redo record will have
     * been disconnected during the checkpoint (so it won't be used).  The
     * checkpoint won't free an invalid entry, so the memory is still there
     * (so this won't core dump).
     */
#ifdef REDO_LOG
    if (nondeterm && redo_flag) {
	if (redo_ptr == NULL) {
	    fprintf(stderr, "error: redo_ptr = NULL\n");
	    exit(1);
	}
	redo_ptr->fd = d;
	redo_ptr->buf = (char *) iov;
	redo_ptr->return_status = status;
	if (status <= 0) {
	    redo_ptr->return_errno = errno;
	    redo_ptr->select_buf_ptr[0] = NULL;
	} else {
	    sel_ptr = redo_ptr->select_buf_ptr[0] =
		vista_malloc(explicit_seg_ptr, sizeof(selbuf), VISTA_NO_TRANS);
	    if (sel_ptr == NULL) {
		fprintf(stderr, "error in readv: vista_malloc of sel_ptr failed\n");
	    }
	    sel_ptr->buf = vista_malloc(explicit_seg_ptr, status,
					VISTA_NO_TRANS);

	    if (sel_ptr->buf == NULL) {
		fprintf(stderr, "error in readv: vista_malloc failed, status = %d\n", status);
		vista_perror("vista_malloc");
		exit(1);
	    }

	    sel_ptr->length = status;

	    /*
	     * Copy input data to redo log record.
	     */
	    copy_from_iov(sel_ptr->buf, status, (struct iovec *) iov, iovcnt);

	    /*
	     * Free this redo record if there was a checkpoint anytime
	     * after the read.  This isn't necessary, but it decreases
	     * memory leakage.
	     */
	    if (explicit_ptr->redo_tail_ptr == &explicit_ptr->redo_head) {
		vista_free(explicit_seg_ptr, sel_ptr->buf, VISTA_NO_TRANS);
		vista_free(explicit_seg_ptr, sel_ptr, VISTA_NO_TRANS);
		vista_free(explicit_seg_ptr, redo_ptr, VISTA_NO_TRANS);
	    }
	}

	/*
	 * This needs to be done AFTER the record is completely initialized.
	 */
	redo_ptr->type = REDO_READ;

#ifndef TRACK_NA
	/*
	 * Flush the redo log to disk (only has effect if using DC_DISK).
	 * This is done if we're not tracking non-abortables, since then we
	 * must assume that the next event is non-abortable.
	 */
	redo_flush();
#endif
    } else if (nondeterm) {
	/*
	 * The redo record wasn't attached to the redo log, so free it.
	 */
	vista_free(explicit_seg_ptr, redo_ptr, VISTA_NO_TRANS);
    }

#else

    /*
     * Reads from non-files (e.g. keyboard) are non-deterministic.
     * I assume that reads from files are deterministic (this assumes there
     * are no shared file accesses, which would be a form of hidden
     * channels).
     */
    if (nondeterm) {
	seen_nd("readv");
    }

#endif /* REDO_LOG */

    return(status);
}

/*
 * Intercept read by transforming it into readv.
 */
ssize_t read(int d, void *buf, size_t nbytes)
{
    struct iovec iov;

    iov.iov_base = buf;
    iov.iov_len = nbytes;

    return(readv(d, &iov, 1));
}

ssize_t writev(int d, const struct iovec *iov, int iovcnt)
{
    struct msghdr msg;
    undo_node *undo_ptr;
    int status;
    int i;
    size_t write_size;	/* size of write data in iov */
    struct stat stat_buf;
    off_t file_size;

    /*
     * Re-direct writev on sockets to sendmsg to take care of connected sockets
     */
    if (explicit_ptr->fds[d].status == FD_SOCKET) {
	msg.msg_name = NULL;
	msg.msg_namelen = 0;
	msg.msg_iov = (struct iovec *) iov; /* typecast to drop const warning */
	msg.msg_control = NULL;
	msg.msg_controllen = 0;
	return(sendmsg(d, &msg, 0));
    }

    /*
     * If the file is regular, make the writev undoable.  Otherwise take a
     * checkpoint (for user files and stdout) (don't checkpoint
     * before writing to stderr, since dc and vista write to stderr).
     * stdout is not marked as a normal file, so writes to stdout will be
     * checkpointed (not logged).
     * Checkpoint before writing to a pipe, because the other process might
     * read the pipe data and so we can't undo it (unless we use 2PC).
     */
    if (d == STDOUT_FILENO || d >= USER_FD) {
	if (explicit_ptr->fds[d].status == FD_FILE &&
		!explicit_ptr->fds[d].u.file.pipe) {

	    /*
	     * Sometimes writev is called by cleanup in dc's exit.  Don't
	     * do any extra work for this request at all.
	     */
	    if (!explicit_ptr->dc_active) {
		fprintf(stderr, "warning: dc's writev called with dc not active\n");
		return(syscall(SYS_writev, d, iov, iovcnt));
	    }

	    /*
	     * Get file size.  I used to try to maintain it dynamically
	     * (e.g. change it on each file write and truncate), but this was
	     * hard because some programs (e.g. nvi) opened a file with two
	     * file descriptors and it was a pain to update the sizes.
	     */
	    fstat(d, &stat_buf);
	    file_size = stat_buf.st_size;

	    prefault();
	    vista_set_range(dc_seg_ptr, &(explicit_ptr->fds[d].u.file.position),
			sizeof(explicit_ptr->fds[d].u.file.position),
			explicit_ptr->dc_tid);
	    /*
	     * Update file position before write operation for O_APPEND files
	     */
	    if (explicit_ptr->fds[d].u.file.flags & O_APPEND) {
		fprintf(stderr, "adjusting position due to O_APPEND\n");
		explicit_ptr->fds[d].u.file.position = file_size;
	    }
	    /*
	     * Log enough information to undo the write.  Log the part
	     * of the write that overwrites existing information, and log
	     * the size so I can truncate the file to the size before the
	     * write operation.
	     * The file may or may not be opened during recovery, because
	     * the open may have occured in the current checkpoint interval.
	     * This prevents me from depending on the recovery process for
	     * open files (e.g. truncating the file to the checkpointed size).
	     */
	    for (i = 0, write_size = 0; i < iovcnt; i++) {
		write_size += iov[i].iov_len;
	    }

	    undo_ptr = vista_malloc(explicit_seg_ptr, sizeof(undo_node),
					VISTA_NO_TRANS);
#ifdef PROG_RETRY
            undo_ptr->chkptid = explicit_ptr->chkpt_id;
#endif /* PROG_RETRY */

	    undo_ptr->type = UNDO_WRITE;
	    undo_ptr->u.write.path = vista_malloc(explicit_seg_ptr,
			    strlen(explicit_ptr->fds[d].u.file.path)+1,
			    VISTA_NO_TRANS);
	    strcpy(undo_ptr->u.write.path, explicit_ptr->fds[d].u.file.path);
	    undo_ptr->u.write.position = explicit_ptr->fds[d].u.file.position;

	    /*
	     * Calculate amount of data to log.  Be careful: file position and
	     * size are off_t (long long).
	     */
	    if (file_size <= explicit_ptr->fds[d].u.file.position) {
		undo_ptr->u.write.datasize = 0;
	    } else if (write_size < (size_t) (file_size -
			    explicit_ptr->fds[d].u.file.position) ) {
		undo_ptr->u.write.datasize = write_size;
	    } else {
		undo_ptr->u.write.datasize = (size_t)
			    (file_size - explicit_ptr->fds[d].u.file.position);
	    }

	    undo_ptr->u.write.filesize = file_size;

	    /* read original data */
	    undo_ptr->u.write.data = vista_malloc(explicit_seg_ptr,
			    undo_ptr->u.write.datasize, VISTA_NO_TRANS);
	    status = syscall(SYS_read, d, undo_ptr->u.write.data,
				undo_ptr->u.write.datasize);
	    if (status != undo_ptr->u.write.datasize) {
		fprintf(stderr, "read of original data (fd %d) (%d bytes) returned %d\n",
		    d, undo_ptr->u.write.datasize, status);
		fprintf(stderr, "path=%s\n", explicit_ptr->fds[d].u.file.path);
		fprintf(stderr, "flags=0x%x\n", explicit_ptr->fds[d].u.file.flags);
		perror("read");
		exit(1);
	    }

	    /* restore file position after reading the original data */
	    sys_lseek(d, explicit_ptr->fds[d].u.file.position, SEEK_SET);

	    undo_ptr->next_ptr = explicit_ptr->undo_head;

	    /*
	     * Add to linked list as last step, so a crash won't result in
	     * a half-initialized undo record being processed.
	     */
	    explicit_ptr->undo_head = undo_ptr;
	} else {
	    /*
	     * This is a non-abortable write.
	     */
	    seen_na("writev");
	}
    }
    status = syscall(SYS_writev, d, iov, iovcnt);

    /*
     * Update file size and position if write succeeded.
     */
    if ( (d == STDOUT_FILENO || d >= USER_FD) &&
	  explicit_ptr->fds[d].status == FD_FILE && status >= 0) {

	prefault();
	vista_set_range(dc_seg_ptr, &(explicit_ptr->fds[d].u.file.position),
		    sizeof(explicit_ptr->fds[d].u.file.position),
		    explicit_ptr->dc_tid);
	explicit_ptr->fds[d].u.file.position += status;
    }

    return(status);
}

/*
 * Intercept write by transforming it into writev.
 */
ssize_t write(int d, const void *buf, size_t nbytes)
{
    struct iovec iov;

    iov.iov_base = (void *) buf;
    iov.iov_len = nbytes;

    return(writev(d, &iov, 1));
}

/*
 * Make a ftruncate system call.  Code taken from
 * /usr/src/lib/libc/sys/ftruncate.c
 * This code is different than a standard syscall because of the 64-bit off_t.
 */
static off_t sys_ftruncate(int d, off_t length)
{
    return(__syscall((quad_t)SYS_ftruncate, d, 0, length));
}

/*
 * Intercept ftruncate to take a checkpoint beforehand (since ftruncate is not
 * yet undoable).
 */
int ftruncate(int d, off_t length)
{
    int status;

    /*
     * I could make ftruncate/truncate save the data that's been truncated,
     * so I don't have to take a checkpoint.
     *
     * This is a non-abortable event, so checkpoint (if there's been
     * a non-deterministic since the last checkpoint).
     */
    seen_na("ftruncate");
    status = sys_ftruncate(d, length);

    return(status);
}

int open(const char *path, int flags, ...)
{
    va_list ap;
    int mode=0;
    int fd;
    struct stat stat_buf;

    va_start(ap, flags);
    if (flags & O_CREAT) {
	mode = va_arg(ap, int);
	va_end(ap);
    }

    /*
     * Need to read the file in order to undo writes.  So open files read/write
     * if the user specific write-only.
     */
    if (flags & O_WRONLY) {
	flags = (flags & ~O_ACCMODE) | O_RDWR;
    }

    fd = syscall(SYS_open, path, flags, mode);

    /*
     * Ignore the following file descriptors:
     *   0: stdin (track the file position if it's a regular file)
     *   1: stdout
     *   2: stderr
     *   3: two-phase commit port (receives messages from 2PC coordinator)
     *   4: <program>.data.meta
     *   5: <program>.data
     *   6: <program>.explicit.meta (also temporarily used by dc to open
     *					<program>.data)
     *   7: <program>.explicit
     *   8: <program>.dc.meta
     *   9: <program>.dc
     *
     * dc doesn't handle programs that redirect stdout/stderr to a file,
     * because when it restarts, the output file will be truncated before
     * dc even starts.  Even if it's redirected with >>, there's a chance
     * that the last write will be successful, yet re-executed after a crash
     * (and hence will appear in the file twice).
     */

    if (fd >= USER_FD) {
	prefault();
	vista_set_range(dc_seg_ptr, &(explicit_ptr->fds[fd]),
			sizeof(explicit_ptr->fds[fd]), explicit_ptr->dc_tid);

	strcpy(explicit_ptr->fds[fd].u.file.path, path);
	explicit_ptr->fds[fd].u.file.flags = flags;
	explicit_ptr->fds[fd].u.file.unlink = 0;
	explicit_ptr->fds[fd].u.file.pipe = 0;

	/*
	 * Save the file position (for repositioning during recovery) for
	 * regular files.
	 */
	fstat(fd, &stat_buf);
	if ( (stat_buf.st_mode & S_IFMT) == S_IFREG) {
	    explicit_ptr->fds[fd].u.file.position = 0;
	    explicit_ptr->fds[fd].select_buf = NULL;
	    explicit_ptr->fds[fd].status = FD_FILE;
	} else {
	    explicit_ptr->fds[fd].select_buf = NULL;
	    explicit_ptr->fds[fd].status = FD_SPECIAL;
	    fprintf(stderr, "warning: dc won't recover file position for %s\n",
		path);
	}
    }

    return(fd);
}

int close(int d)
{
    if (d >= USER_FD) {
	if (explicit_ptr->fds[d].status == FD_FILE &&
		explicit_ptr->fds[d].u.file.unlink) {
	    syscall(SYS_unlink, explicit_ptr->fds[d].u.file.path);
	}
	prefault();
	vista_set_range(dc_seg_ptr, &(explicit_ptr->fds[d].status),
			sizeof(explicit_ptr->fds[d].status),
			explicit_ptr->dc_tid);
	explicit_ptr->fds[d].status = FD_CLOSED;

	if (explicit_ptr->fds[d].select_buf != NULL) {
	    /*
	     * Free the pending select buffer.
	     */
	    vista_free(explicit_seg_ptr, explicit_ptr->fds[d].select_buf->buf,
			explicit_ptr->dc_tid);

	    vista_set_range(dc_seg_ptr, &(explicit_ptr->fds[d].select_buf),
			    sizeof(explicit_ptr->fds[d].select_buf),
			    explicit_ptr->dc_tid);
	    explicit_ptr->fds[d].select_buf = NULL;

	}
    }
    return(syscall(SYS_close, d));
}

/*
 * Trap dup and dup2 (just to issue warning messages).  I could probably get
 * DC to recover this if I really needed to.
 */
int dup(int oldd)
{
    fprintf(stderr, "warning: DC won't recover dup'ed file descriptors\n");
    return(syscall(SYS_dup, oldd));
}

int dup2(int oldd, int newd)
{
    fprintf(stderr, "warning: DC won't recover dup2'ed file descriptors\n");
    return(syscall(SYS_dup2, oldd, newd));
}

/*
 * Trap pipe by creating and opening (twice) a normal file.  This allows
 * me to re-open it during recovery.  Otherwise it's not possible to
 * recover the same pipe (which is still open by the surviving process).
 * Eventually it would be nice to make this a circular queue instead of
 * an ever-growing one (could store the head and tail pointers in the first 2
 * words of the file; or could have producer block after file reaches
 * a certain size, then wait for consumer to truncate the file).
 */
int pipe(int *fildes)
{
    int i;
    char name[PATH_MAX];
    int success;

    /*
     * Try to create the pipe file (allow up to 10 pipes per process).
     */
    for (i=0, success=0; i<10 && !success; i++) {
	sprintf(name, "%s.%s.pipe.%d", DC_FILE, dc_command, i);

	fildes[0] = open(name, O_RDONLY | O_CREAT | O_EXCL, 0777);

	if (fildes[0] >= 0) {
	    success = 1;
	} else if (errno != EEXIST) {
	    perror("pipe open");
	    exit(1);
	}
    }
    if (!success) {
	fprintf(stderr, "pipe couldn't find a new pipe file\n");
	exit(1);
    }

    fildes[1] = open(name, O_WRONLY);
    if (fildes[1] < 0) {
	perror("pipe open");
	exit(1);
    }

    /*
     * Mark the file descriptors as pipes, so DC knows to checkpoint before
     * each write to that descriptor.
     */
    prefault();
    for (i=0; i<2; i++) {
	vista_set_range(dc_seg_ptr, &(explicit_ptr->fds[fildes[i]].u.file.pipe),
		sizeof(explicit_ptr->fds[fildes[i]].u.file.pipe),
		explicit_ptr->dc_tid);
	explicit_ptr->fds[fildes[i]].u.file.pipe = 1;
    }

    return(0);
}

/*
 * Make it look like the recovered process has the same pid as the original
 * process.  Remember the parent pid in case the parent dies and recovers.
 */
pid_t getpid()
{
    return(original_pid);
}

pid_t getppid()
{
    return(original_ppid);
}

/*
 * Maintain mapping from original pid to current pid (stored in a file)
 */
set_current_pid()
{
    int fd;
    int current_pid;

    current_pid = syscall(SYS_getpid);

    /*
    fprintf(stderr, "writing current pid %d to pid table for original pid %d\n",
	current_pid, original_pid);
    */

    fd = syscall(SYS_open, "/usr/tmp/dc.pid", O_WRONLY|O_CREAT, 0777);
    if (fd < 0) {
	fprintf(stderr, "set_current_pid couldn't open /usr/tmp/dc.pid\n");
	exit(1);
    }
    /*
     * Make the pid table writable by all (in case the above open just created
     * the file).
     */
    fchmod(fd, 0777);

    sys_lseek(fd, original_pid * sizeof(original_pid), SEEK_SET);
    syscall(SYS_write, fd, &current_pid, sizeof(current_pid));
    syscall(SYS_close, fd);
}

/*
 * Return current pid for the specified original pid.
 */
pid_t get_current_pid(pid_t pid)
{
    int fd;
    int current_pid;

    fd = syscall(SYS_open, "/usr/tmp/dc.pid", O_RDONLY);
    if (fd < 0) {
	current_pid = pid;
    } else {
	sys_lseek(fd, pid * sizeof(pid), SEEK_SET);
	if (syscall(SYS_read, fd, &current_pid, sizeof(current_pid)) !=
		sizeof(current_pid) || current_pid == 0) {
	    /*
	     * Table didn't have an entry for this process.  DC only fills in
	     * the table when current pid != original pid.  Or this could be a
	     * non-DC process.
	     */
	    current_pid = pid;
	}
    }

    syscall(SYS_close, fd);
    /*
    fprintf(stderr, "get_current_pid returning current pid %d for original pid %d\n",
	current_pid, pid);
    */

    return(current_pid);
}

/*
 * Trap kill, so we can send signal to current pid.
 */
int kill(pid_t pid, int sig)
{
    pid_t current_pid;

    /* fprintf(stderr, "asked to send signal to process %d\n", pid); */

    current_pid = get_current_pid(pid);

    return(syscall(SYS_kill, current_pid, sig));
}

/*
 * Wrapper functions around Vista calls.  DC versions of libc functions.
 */
void *malloc(size_t size)
{
    void *status;

    /*
    write(2, "dc_malloc\n", 10);
    */

    /*
     * Some user recovery routines (invoked by dc_recover) will call malloc
     * and free (the most notable example are curses setup routines.  This
     * is a problem, since it may change the state of the program from that
     * saved during the checkpoint.  I don't have a fix for this.  For now,
     * just do the malloc/free NO_TRANS and hope that the user recovery
     * routine frees all allocated data (or that it doesn't allocate new buckets
     * or otherwise change replay). ???
     */

    if (! explicit_ptr->dc_active) {
	status = vista_malloc(dc_seg_ptr, size, VISTA_NO_TRANS);
	fprintf(stderr, "warning: dc's malloc called when dc isn't active	0x%x\n", status);
	return(status);
    }

    prefault();
    status = vista_malloc(dc_seg_ptr, size, explicit_ptr->dc_tid);

    /*
    fprintf(stderr, "malloc(%d)\n", size);
    */

    return(status);
}

void free(void *ptr)
{
    /*
    write(2, "dc_free\n", 8);
    */

    /*
     * See note in malloc().
     */
    if (! explicit_ptr->dc_active) {
	fprintf(stderr, "warning: dc's free called when dc isn't active	0x%x\n", ptr);
	vista_free(dc_seg_ptr, ptr, VISTA_NO_TRANS);
	return;
    }

    prefault();
    vista_free(dc_seg_ptr, ptr, explicit_ptr->dc_tid);
}

void *realloc(void *ptr, size_t size)
{
    void *new;
    /* write(2, "dc_realloc\n", 11); */
    new = malloc(size);

    if (ptr) {
	memcpy(new, ptr, size);
	free(ptr);
    }
    return(new);
}

/*
 * Intercept fprintf, so we can do special stuff for fprintf(stderr, ...).
 * Use sprintf/write to avoid getting prot faults when fprintf is called from
 * within a Vista function.  Don't print lines longer than 400 characters
 * from within Vista.  This is really just a programming convenience for Vista
 * implementors.
 */
int fprintf(FILE *fp, const char *fmt, ...)
{
    int ret;
    va_list ap;
    FILE f;
    char buf[400];

    va_start(ap, fmt);

#ifdef DEBUG
    /*
     * It's safer when debugging to use sprintf/write.  Plus, sometimes
     * things get messed up when mixing write and buffered I/O.
     */
    if (fp == stderr)
#else
    if (fp == stderr && unrecov_ptr && (lock_v || lock_p) )
#endif
	{
	/*
	 * This code comes from /usr/src/lib/libc/stdio/sprintf.c
	 * Calling sprintf would have been easier, but sprintf doesn't take
	 * a va_list as the third argument.
	 */
        f._file = -1;
        f._flags = __SWR | __SSTR;
        f._bf._base = f._p = (unsigned char *)buf;
        f._bf._size = f._w = INT_MAX;
        ret = vfprintf(&f, fmt, ap);
        va_end(ap);
        *f._p = 0;
	write(STDERR_FILENO, buf, strlen(buf));
	return(strlen(buf));
    } else {
	ret = vfprintf(fp, fmt, ap);
	va_end(ap);
	return (ret);
    }
}

int gettimeofday(struct timeval *tp, struct timezone *tzp)
{
    int status;

    dc_lock();

    dc_touch((char *) tp, sizeof(*tp));
    dc_touch((char *) tzp, sizeof(*tzp));
    status = syscall(SYS_gettimeofday, tp, tzp);

    dc_unlock();

    seen_nd("gettimeofday");

    return(status);
}

int getdirentries(int fd, char *buf, int nbytes, long *basep)
{
    int status;
    dc_lock();

    dc_touch(buf, nbytes);
    status = syscall(SYS_getdirentries, fd, buf, nbytes, basep);

    dc_unlock();

    return(status);
}

int stat(const char *path, struct stat *sb)
{
    int status;

    dc_lock();
    dc_touch((char *) sb, sizeof(*sb));
    status = syscall(SYS_stat, path, sb);

    dc_unlock();

    return(status);
}

int lstat(const char *path, struct stat *sb)
{
    int status;

    dc_lock();
    dc_touch((char *) sb, sizeof(*sb));
    status = syscall(SYS_lstat, path, sb);

    dc_unlock();

    return(status);
}

int fstat(int fd, struct stat *sb)
{
    int status;

    /*
     * vista and DC use fstat--make sure these calls don't go back into vista.
     * vista shouldn't need dc_touch, because it only uses stack variables.
     */
    if (fd >= USER_FD) {
	dc_lock();
	dc_touch((char *) sb, sizeof(*sb));
	status = syscall(SYS_fstat, fd, sb);
	dc_unlock();
    } else {
	status = syscall(SYS_fstat, fd, sb);
    }
    return(status);
}

/* Trap __sysctl (sysctl calls __sysctl) */
int __sysctl(int *name, u_int namelen, void *oldp, size_t *oldlenp, void *newp,
             size_t newlen)
{
    int status;

    dc_lock();
    dc_touch(oldp, *oldlenp);
    status = syscall(SYS___sysctl, name, namelen, oldp, oldlenp, newp, newlen);
    dc_unlock();

    return(status);
}

/*
 * Setitimer.  Restore the most recent timer value during recovery.  This may
 * lead to an extra timer interrupt if the timer was set to do a one-shot,
 * and the one-shot timer happened.  When we recover, the one-shot will be
 * restored.
 */
int setitimer(int which, const struct itimerval *value,
		struct itimerval *ovalue)
{
    if (which != ITIMER_REAL) {
	fprintf(stderr, "error: setitimer, with which = %d\n", which);
	exit(1);
    }
    timer = *value;
    return(syscall(SYS_setitimer, which, value, ovalue));
}


/*
 * sigaction:
 *   save the signal handler (so that dc can set the handlers during recovery)
 *   and call vista_sigaction.
 */
int sigaction(int sig, const struct sigaction *act, struct sigaction *oact)
{
    struct sigaction act1;
    int status;

    if (sig < 1 || sig >= NSIG) {
	fprintf(stderr, "dc's sigaction: sig=%d\n", sig);
	exit(1);
    }

    /* fprintf(stderr, "dc's sigaction called for signal %d\n", sig); */

    if (act != NULL) {

	if (act->sa_handler == vista_sigprot ||
	    act->sa_handler == vista_sigother || act->sa_handler == SIG_DFL) {

	    /*
	     * This is being called by Vista (or the handler is SIG_DFL), so
	     * just do the syscall.
	     * Remember that Vista doesn't use SIG_DFL for sigprot.

	    /*
	    fprintf(stderr, "dc's sigaction got Vista's sigaction call for signal %d\n", sig);
	    */
	    act1 = *act;

	    act1.sa_handler = dc_sighandler;

	    /*
	     * DC implements the SA_RESTART functionality, so don't officially
	     * register the signal as SA_RESTART.
	     */
	    act1.sa_flags &= (~SA_RESTART);

	    /*
	     * Don't let the user mask sigprot or SIG2PC.
	     */
	    sigdelset(&act1.sa_mask, VISTA_SIGPROT);
	    sigdelset(&act1.sa_mask, SIG2PC);

	    /*
	     * All signals must use alternate (explicitly logged) stack.
	     * Otherwise the system's code to deliver a signal to the process
	     * might encounter a write-protected page (on the normal stack).
	     * This causes a SIGILL.
	     */
	    act1.sa_flags |= SA_ONSTACK;

	    return(syscall(SYS_sigaction, sig, &act1, oact));

	} else {
	    /* save a copy for recovery */
	    /*
	    fprintf(stderr, "dc's sigaction saving handler for signal %d\n",
		    sig);
	    */
	    dc_siglist[sig] = act1 = *act;
	    /*
	    fprintf(stderr, "setting dc_siglist[%d].sa_handler = 0x%x\n",
		sig, dc_siglist[sig].sa_handler);
	    */

	    /*
	     * Re-direct to dc_sighandler2.
	     * So the complete order of signal handlers is:
	     *     dc_sighandler -> vista_sigother -> dc_sighandler2 ->
	     *     user's signal handler
	     */
	    act1.sa_handler = dc_sighandler2;

	    /* fprintf(stderr, "dc's sigaction calling vista_sigaction\n"); */
	    if (sig == VISTA_SIGPROT) {
		/*
		 * Don't use dc_sighandler2 for SIGPROT.  It's not needed,
		 * because SIGPROT is deterministic.  Plus it breaks, because
		 * it doesn't have the parameters it needs to call the
		 * user's SIGPROT handler, e.g. the address.
		 */
		status = vista_sigaction(sig, act, oact);
	    } else {
		status = vista_sigaction(sig, &act1, oact);
	    }
	    return(status);
	}
    }

    /* fprintf(stderr, "dc's sigaction calling vista_sigaction with NULL\n"); */
    status = vista_sigaction(sig, act, oact);

    return(status);
}

/*
 * My version of sigprocmask.  Don't let the user mask sigprot or SIG2PC.
 */
int sigprocmask(int how, const sigset_t *set, sigset_t *oset)
{
    sigset_t set1;

    if (set != NULL && how != SIG_UNBLOCK) {
	set1 = *set;
	sigdelset(&set1, VISTA_SIGPROT);
	sigdelset(&set1, SIG2PC);
	return(sys_sigprocmask(how, &set1, oset));
    } else {
	return(sys_sigprocmask(how, set, oset));
    }
}

void dc_init_internal(char *command, void (*check_ptr)(), void (*recov_ptr)(),
    char ***argvp, char ***envpp)
{
    /*
     * Discount Checking's data layout in Vista space:
     *
     *    data_seg: implicitly logged
     *		data (globals, bss, dc_stack, etc.)
     *		mapped on top of original data segment
     *
     *    explicit_seg: explicitly logged, see explicit_type struct in dc.h
     *		signal stack (fixed size)
     *		miscellaneous variables
     *		copy of Vista's unrecovered data
     *
     *    dc_seg: vista's user heap, implicitly logged
     */

    int fd;
    char *addr;
    long stack_size;
    long data_size;
    char name[PATH_MAX];
    int i;
    struct stat stat_buf;
    sigset_t set;
    struct sigaction act;
#ifdef TWO_PHASE_COMMIT
    int child_pid;
    msg message;
    int got_registered;
    struct timeval timeout = {1, 0};
    fd_set fdset;
#endif
    extern char **environ;

    syscall(SYS_gettimeofday, &t_start, NULL);

    /*
     * ???
     * show which file descriptors are open
    for (fd=0; fd<16; fd++) {
	int status;
	status = syscall(SYS_fstat, fd, &stat_buf);
	if (status < 0) {
	    fprintf(stderr, "fd %d is closed\n", fd);
	} else {
	    fprintf(stderr, "fd %d is open, type is %o, position is %d\n", fd,
		stat_buf.st_mode & S_IFMT, sys_lseek(fd, 0, SEEK_CUR));
	}
    }
     */

    /*
     * Open 2PC socket
     */
    socket_2pc = socket(AF_INET, SOCK_DGRAM, 0);
    if (socket_2pc < 0) {
	perror("socket");
	exit(1);
    }

    vista_init();

    for (i=1; i<NSIG; i++) {
	dc_siglist[i].sa_handler = SIG_DFL;
    }

    timer.it_interval.tv_sec = timer.it_interval.tv_usec =
	timer.it_value.tv_sec = timer.it_value.tv_usec = 0;

#ifdef TWO_PHASE_COMMIT
    /*
     * Block SIG2PC until we're ready to get prepare signals.
     */
    sigemptyset(&set);
    sigaddset(&set, SIG2PC);
    sys_sigprocmask(SIG_BLOCK, &set, NULL);
    /*
     * Fork a SIG2PC daemon to send 2PC prepare signals to the main process.
     * The child process is the main process; the parent process is the
     * SIG2PC daemon.
     */
    if (child_pid = fork()) {
	/*
	 * Open port to listen for messages from the 2PC coordinator
	 */
	int sig_sock;
	struct sockaddr_in name;
	msg m_buf;
	struct sigaction act;

	/*
	 * Exit if the main process dies.
	 */
	act.sa_handler = _exit;
	act.sa_mask = 0;
	act.sa_flags = SA_NOCLDSTOP;
	syscall(SYS_sigaction, SIGCHLD, &act, NULL);

	sig_sock = syscall(SYS_socket, AF_INET, SOCK_DGRAM, 0);
	if (sig_sock < 0) {
	    perror("socket");
	    exit(1);
	}

	bzero(&name, sizeof(name));
	name.sin_family = AF_INET;
	name.sin_addr.s_addr = htonl(INADDR_ANY);
	name.sin_port = htons(SIG_PORT);

	while( syscall(SYS_bind, sig_sock, (struct sockaddr *) &name,
		sizeof(name)) < 0) {
	    perror("bind");

	    /*
	     * An old SIG2PC daemon may be still alive.  It will die when it
	     * tries to send a signal to the deceased child; then the new
	     * SIG2PC daemon will bind successfully.
	     */
	    sleep(1);
	}

	while (1) {
	    syscall(SYS_recvfrom, sig_sock, &m_buf, sizeof(m_buf), 0, NULL,
		    NULL);
	    fprintf(stderr, "SIG2PC daemon: (%d) sending SIG2PC to process %d\n",
		m_buf.id, child_pid);
	    if (syscall(SYS_kill, child_pid, SIG2PC)) {
		fprintf(stderr, "SIG2PC daemon: kill failed; exiting\n");
		_exit(1);
	    }
	}
    }

#endif

    /*
     * are we starting a fresh process, or are we recovering?
     */
    if (fresh_process(command)) {
	fprintf(stderr, "Starting fresh process.  Switching to Vista space.\n");
    } else {
	fprintf(stderr, "recovering\n");
	dc_recover(command, 1);
    }

    /*
     * Map vista space over data segment.  The order of these operations
     * is crucial.  Have as little as possible between copying the data
     * segment into the file and vista_malloc.
     */
    
    sprintf(name, "%s.%s.data", DC_FILE, command);
    data_seg_ptr = vista_map(name, DATA_START);
    if (data_seg_ptr == NULL) {
	vista_perror("vista_map of data segment failed");
	exit(1);
    }

    /*
     * Copy the current data segment to vista's data file.
     * Don't change any variables in data segment from here until after
     * vista_malloc, because vista_malloc will overwrite them with the saved
     * values.  Data segment goes from DATA_START to sbrk(0).
     */

    data_size = (char *) sbrk(0) - DATA_START;
    /* might need to round data_size up to nearest page */

    sprintf(name, "%s.%s.data", DC_FILE, command);
    fd = open(name, O_RDWR, 0777);
    if (fd < 0) {
	perror("open");
	exit(1);
    }

    if (write(fd, DATA_START, data_size) != data_size) {
	perror("write");
	exit(1);
    }

    addr = vista_malloc(data_seg_ptr, data_size, VISTA_NO_TRANS);
    if (addr != DATA_START) {
	fprintf(stderr, "error: vista_malloc of data segment returned %lx\n",
		addr);
	exit(1);
    }

    close(fd);

    /*
     * map explicit segment
     */
    sprintf(name, "%s.%s.explicit", DC_FILE, command);
    explicit_seg_ptr = vista_map(name, NULL);
    if (explicit_seg_ptr == NULL) {
	vista_perror("vista_map of explicit segment failed");
	exit(1);
    }

    /* allocate explicitly logged segment */
    explicit_ptr = vista_malloc(explicit_seg_ptr, sizeof(explicit_type),
				VISTA_NO_TRANS);
    explicit_ptr->dc_active = 0;

    for (i=0; i<OPEN_MAX; i++) {
	explicit_ptr->fds[i].status = FD_CLOSED;
    }

    explicit_ptr->undo_head = NULL;
    explicit_ptr->num_checkpoints = 0;

#ifdef TWO_PHASE_COMMIT
    explicit_ptr->initiate_abort = explicit_ptr->sent_message = 0;
#endif

    /*
     * stdin is special.  We want to save and recover its file position
     * if it's a regular file.
     */
    fstat(STDIN_FILENO, &stat_buf);
    if ( (stat_buf.st_mode & S_IFMT) == S_IFREG) {
	explicit_ptr->fds[STDIN_FILENO].status = FD_FILE;
	explicit_ptr->fds[STDIN_FILENO].select_buf = NULL;
	strcpy("", explicit_ptr->fds[STDIN_FILENO].u.file.path);
	explicit_ptr->fds[STDIN_FILENO].u.file.flags = 0;
	explicit_ptr->fds[STDIN_FILENO].u.file.unlink = 0;
	explicit_ptr->fds[STDIN_FILENO].u.file.pipe = 0;
	explicit_ptr->fds[STDIN_FILENO].u.file.position = 0;
    } else {
	explicit_ptr->fds[STDIN_FILENO].status = FD_SPECIAL;
	explicit_ptr->fds[STDIN_FILENO].select_buf = NULL;
    }

    /*
     * Initialize our copy of Vista's unrecovered data, then point
     * Vista's unrecov_ptr at our copy (in the explicit segment).
     */
    memcpy(&(explicit_ptr->unrecov_data), unrecov_ptr, sizeof(unrecov));
    unrecov_ptr = &(explicit_ptr->unrecov_data);

    /*
     * switch to dc_stack (persistent)
     *
     * original stack goes from [$sp, STACK_START)
     * dc_stack goes from
     *   [dc_stack+MAX_STACK_SIZE-stack_size, dc_stack+MAX_STACK_SIZE)
     */

    if (!setjmp(explicit_ptr->context)) {
	stack_size = STACK_START - SP(explicit_ptr->context);

	if (stack_size > MAX_STACK_SIZE) {
	    fprintf(stderr, "error: stack too large\n");
	    exit(1);
	}
	if (stack_size < 0) {
	    fprintf(stderr, "error: stack shrunk above STACK_START\n");
	    exit(1);
	}

	/*
	 * initialize dc_stack.  Using memcpy here makes sure it's linked
	 * before vista_sigprot uses memcpy (this was before, when I was trying
	 * to support dynamically linked libraries).  Linking in memcpy during
	 * vista_sigprot causes a prot fault when it writes to the global
	 * offset table (in the data segment, which is protected).
	 */
	memcpy(dc_stack+MAX_STACK_SIZE-stack_size, SP(explicit_ptr->context),
	    stack_size);

	/*
	 * switch to dc_stack
	 */
	if (dc_stack > STACK_START) {
	    fprintf(stderr, "dc_stack at 0x%lx, STACK_START at 0x%lx\n",
		dc_stack, STACK_START);
	    exit(1);
	}
	SP(explicit_ptr->context) -= STACK_START - (dc_stack + MAX_STACK_SIZE);
	FP(explicit_ptr->context) -= STACK_START - (dc_stack + MAX_STACK_SIZE);

	longjmp(explicit_ptr->context, 1);

	/* this is never reached */
    }

    /*
     * Now we're on the dc_stack.
     */

    set_signal_stack(explicit_ptr->sig_stack, 0);

    user_check_ptr = check_ptr;
    user_recov_ptr = recov_ptr;

    strcpy(dc_command, command);

    /*
     * create dc heap
     */
    sprintf(name, "%s.%s.dc", DC_FILE, command);
    dc_seg_ptr = vista_map(name, NULL);
    if (dc_seg_ptr == NULL) {
	vista_perror("vista_map of dc segment failed");
	exit(1);
    }

#ifdef PROG_RETRY
    explicit_ptr->head = 0;
    explicit_ptr->tail = 0;
    explicit_ptr->lastretry = -1;
    explicit_ptr->lastcrash = -1;
    explicit_ptr->chkpt_id = 0;
#endif /* PROG_RETRY */


    /*
     * Tell Vista to put all implicitly logged data into dc_seg
     */
    vista_set_log(dc_seg_ptr);

    explicit_ptr->dc_tid = vista_begin_transaction(dc_seg_ptr, VISTA_IMPLICIT);

    /*
     * Start the transaction on the data segment.  The undo records for the
     * data segment are stored in dc_seg, so that commit of the dc_seg
     * transaction atomically commits the transactions in dc_seg, data_seg,
     * and explicit_seg.
     * The transaction on the dc segment must surround (ie. begin before
     * and end after) the transaction on the data segment, because prot
     * faults when writing to the data segment must be able to store into
     * an active transaction in the dc segment.
     */
    explicit_ptr->data_tid =
	vista_begin_transaction(data_seg_ptr, VISTA_IMPLICIT);

#ifdef DEBUG
    /*
     * Verify that USER_FD is accurate.  ??? This doesn't account for
     * file descriptors inherited from a parent process.
     */
    fd = syscall(SYS_open, "/dev/null", O_RDWR);
    if (fd != USER_FD) {
	fprintf(stderr, "error: first user file descriptor (%d) != %d\n",
	    fd, USER_FD);
	exit(1);
    } else {
	close(fd);
    }
#endif

    /*
     * Set the close-on-exec flag for DC file descriptors.
     */
    for (fd = 0; fd < USER_FD; fd++) {
	if (fd != STDIN_FILENO && fd != STDOUT_FILENO && fd != STDERR_FILENO) {
	    syscall(SYS_fcntl, fd, F_SETFD, 1);
	}
    }

    /*
     * Adjust argv, envp (and the pointers in their arrays) to point them
     * to new stack.
     */
    if (argvp != NULL) {
	argvp = (char ***)
		((char *) argvp - (STACK_START - (dc_stack + MAX_STACK_SIZE)));
	*argvp = (char **)
	    ((char *) (*argvp) - (STACK_START - (dc_stack + MAX_STACK_SIZE)));

	for (i=0; (*argvp)[i] != NULL; i++) {
	    (*argvp) [i] -= STACK_START - (dc_stack + MAX_STACK_SIZE);
	}
    }

    environ = (char **)
	((char *) environ - (STACK_START - (dc_stack + MAX_STACK_SIZE)));
    for (i=0; environ[i] != NULL; i++) {
	environ[i] -= STACK_START - (dc_stack + MAX_STACK_SIZE);
    }
    if (envpp != NULL) {
	envpp = (char ***)
		((char *) envpp - (STACK_START - (dc_stack + MAX_STACK_SIZE)));
	*envpp = (char **)
	    ((char *) (*envpp) - (STACK_START - (dc_stack + MAX_STACK_SIZE)));
    }

    /*
     * Don't buffer for stdin (for ttys).  Otherwise getc issues a read(64KB)
     * (but only returns what the user actually typed), which causes a
     * big dc_touch for each input.
     * dc_touch is needed here because I've just protected the entire stack
     * (which is where stat_buf lives).
     */
    dc_touch( (char *) &stat_buf, sizeof(stat_buf));
    fstat(STDIN_FILENO, &stat_buf);
    if ( (stat_buf.st_mode & S_IFMT) != S_IFREG) {
	setbuf(stdin, NULL);
    }

    /*
     * Remember the original pid and parent's pid
     */
    original_pid = syscall(SYS_getpid);
    original_ppid = syscall(SYS_getppid);
    /*
    fprintf(stderr, "original pid = %d, original ppid = %d\n", original_pid,
	original_ppid);
    */

    /*
     * Initialize the redo log.
     */
    explicit_ptr->redo_head.next_ptr = NULL;
    explicit_ptr->redo_tail_ptr = &explicit_ptr->redo_head;
    explicit_ptr->redo_lastflush_ptr = &explicit_ptr->redo_head;

#ifdef TWO_PHASE_COMMIT
    /*
     * Register with the server as part of the 2PC group
     */
    fprintf(stderr, "registering with 2PC coordinator\n");
    bzero(&serv_addr, sizeof(serv_addr));
    serv_addr.sin_family = AF_INET;
    serv_addr.sin_addr.s_addr = SERVER_IP;
    serv_addr.sin_port = htons(SERVER_PORT);

    do {
	/* send register message */
	message.type = REGISTER;

	if (syscall(SYS_sendto, socket_2pc, &message, sizeof(message), 0,
		(struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0) {
	    perror("sendto");
	    exit(1);
	}

	FD_ZERO(&fdset);
	FD_SET(socket_2pc, &fdset);

	got_registered = 0;
	if (syscall(SYS_select, OPEN_MAX, &fdset, NULL, NULL, &timeout) == 1) {
	    if (syscall(SYS_recvfrom, socket_2pc, &message, sizeof(message),
		    0, NULL, NULL) == sizeof(message)) {
		if (message.type == REGISTERED) {
		    got_registered = 1;
		    explicit_ptr->interval_2pc = message.new_id;
		}
	    }
	}
    } while (!got_registered);

    /*
     * Install the SIG2PC handler.  Use sigaction so this gets re-installed
     * during recovery, and so SIG2PC signals are deferred while the vista_lock
     * is held
     */
    act.sa_handler = dc_checkpoint;
    sigemptyset(&act.sa_mask);
    act.sa_flags = SA_NODEFER | SA_RESTART;
    sigaction(SIG2PC, &act, NULL);

    /*
     * Enable SIG2PC
     */
    sigemptyset(&set);
    sigaddset(&set, SIG2PC);
    sys_sigprocmask(SIG_UNBLOCK, &set, NULL);
#endif

#ifdef DC_DISK
    /*
     * Apparently, the first call to fsync writes to some variable in the
     * data segment (I have no idea why).  The first fsync is done
     * in the middle of Vista's commit_transaction, which holds Vista's locks.
     * This causes a re-entry into Vista's prot handler, which doesn't like
     * getting called with Vista's locks being held.  To get around this,
     * I call fsync here--this pre-faults the variable while it's still safe.
     * Another fix is to change vista to use syscall(SYS_fsync).
     */
    fsync(-1);
#endif

    /*
     * ???
     * I often run performance tests with output piped to a file.  Make
     * sure the results are as if stdout was to a tty.
     */
    setvbuf(stdout, stdout_buf, _IOLBF, 1024);
    setbuf(stderr, NULL);

    /* ???
    {
	int i;
	int buf[4096];
	fprintf(stderr, "pre-allocating 2000 pages in undo log\n");
	for (i=0; i<2000; i++) {
	    vista_set_range(dc_seg_ptr, buf, 4096, explicit_ptr->dc_tid);
	}
    }
    */

    explicit_ptr->dc_active = 1;
}

void dc_checkpoint()
{
    char *sp;
    char *exp_start;
    sigset_t set;
    int fd;
    int pos;
    struct stat stat_buf;
    undo_node *undo_free_ptr, *tmp_undo_ptr;
    int last_tid;
    int recovering;
    int commit;
#ifdef PROG_RETRY
    int chkpttail, chkpthead, nchk;             
#endif /* PROG_RETRY */

#ifdef TWO_PHASE_COMMIT
    fd_set fdset;
    struct timeval timeout = {1, 0};
    int got_vote;
    char buf;
    msg message;
#endif
    redo_node *redo_ptr, *next_redo_ptr;

    if (!explicit_ptr->dc_active) {
	fprintf(stderr, "error: dc_checkpoint called with dc_active = 0\n");
	return;
    }

    prefault();

    /*
     * Turn off signals to prevent a signal coming in between vista calls
     * during dc_checkpoint.  Also use the sigprocmask to remember the current
     * signal mask, so it can be restored during recovery.
     *
     * Note that SIG2PC is also masked here.  This prevents the situation
     * where I'm in the middle of doing a 2-phase commit, and another 2pc
     * request comes in from the 2pc coordinator (possible if multiple
     * processes initiate a two-phase commit at the same time).  The small
     * window when I'm in dc_checkpoint but haven't yet masked signals is
     * no problem, because dc_checkpoint hasn't yet done anything up to this
     * point.  If a second 2pc request comes in, I'll handle that one, then
     * continue doing the one that got interrupted.
     */
    sigfillset(&set);
    sigdelset(&set, VISTA_SIGPROT);

    /* allow ctrl-C, so you can manually kill the process during a checkpoint */
    sigdelset(&set, SIGINT);
    sigdelset(&set, SIGABRT);
    sys_sigprocmask(SIG_BLOCK, &set, &(explicit_ptr->sigmask));

    /*
    fprintf(stderr, "------- Taking a checkpoint -------\n");
    */
    if (explicit_ptr->num_checkpoints == 0) {
	total_bytes_logged = 0;
    }
    explicit_ptr->num_checkpoints++;

    /* ??? calculate rate at which checkpoints are being taken
    if (! (explicit_ptr->num_checkpoints % 100)) {
	struct timeval now_time;
	syscall(SYS_gettimeofday, &now_time, NULL);
	fprintf(stderr, "time for last 100 checkpoints: %lf seconds\n",
	    (now_time.tv_sec - explicit_ptr->last_time.tv_sec) +
	    (now_time.tv_usec - explicit_ptr->last_time.tv_usec) / 1000000.0);
	explicit_ptr->last_time = now_time;
    }
    */

    /*
     * Call any user-specified checkpointing function.  Don't allow auto
     * checkpoints in user's checkpoint function.
     */
    if (user_check_ptr != NULL) {
	explicit_ptr->dc_active = 0;
	(*user_check_ptr)();
	explicit_ptr->dc_active = 1;
    }

    /*
     * Save floating point state.
     */
    float_save();

#ifdef TWO_PHASE_COMMIT
    /*
     * Clear the sent bit.  This needs to be atomic with the prepare,
     * because if I crash after the bit is cleared but before prepare, the
     * sent messages need to be aborted.
     */
    vista_set_range(dc_seg_ptr, &(explicit_ptr->sent_message),
		    sizeof(explicit_ptr->sent_message), explicit_ptr->dc_tid);
    explicit_ptr->sent_message = 0;
#endif

    /*
     * Save the current tid for the DC segment (used in two-phase commit).
     */
    last_tid = explicit_ptr->dc_tid;

    /*
     * Verify that I've maintained file position correctly.
     */
#ifdef DEBUG
    for (fd=0; fd<OPEN_MAX; fd++) {
	if (explicit_ptr->fds[fd].status == FD_FILE) {

	    pos = sys_lseek(fd, 0, SEEK_CUR);
	    if (explicit_ptr->fds[fd].u.file.position != pos) {
		fprintf(stderr, "error: saved position (%d) != real position (%d)\n",
		    explicit_ptr->fds[fd].u.file.position, pos);
		exit(1);
	    }
	}
    }
#endif

    /*
     * Delete DC's undo log.  Setting undo_head to NULL must be done atomically
     * with the end of the transaction.  Upon abort, the undo log should have
     * all the undo records for the current transaction.
     * The undo records are freed after the transaction ends.
     */

    /* save head pointer for use after transaction ends */
#ifndef PROG_RETRY
    undo_free_ptr = explicit_ptr->undo_head;

    vista_set_range(dc_seg_ptr, &(explicit_ptr->undo_head),
		    sizeof(explicit_ptr->undo_head), explicit_ptr->dc_tid);
    explicit_ptr->undo_head = NULL;
#endif /* PROG_RETRY */

    /*
     * Truncate DC's redo log.  This must be done atomically with the
     * checkpoint that's being taken, so do the truncation TRANS.  If a crash
     * occurs before this checkpoint is actually taken, then we need to
     * preserve the redo log (i.e. NOT truncate it).
     */
    for (redo_ptr = explicit_ptr->redo_head.next_ptr; redo_ptr != NULL;
	    redo_ptr = next_redo_ptr) {

	/*
	 * Remember the next record after redo_ptr, since we'll be freeing the
	 * record at redo_ptr and will lose access to redo_ptr->next_ptr.
	 */
	next_redo_ptr = redo_ptr->next_ptr;

	/*
	 * Free record at redo_ptr.  Don't free invalid records, since they
	 * are in the process of being filled in.
	 */
	if (redo_ptr->type == REDO_RECEIVE || redo_ptr->type == REDO_READ) {
	    if (redo_ptr->select_buf_ptr[0] != NULL) {
		vista_free(explicit_seg_ptr, redo_ptr->select_buf_ptr[0]->buf,
			    explicit_ptr->dc_tid);
		vista_free(explicit_seg_ptr, redo_ptr->select_buf_ptr[0],
			    explicit_ptr->dc_tid);
	    }

	vista_free(explicit_seg_ptr, redo_ptr, explicit_ptr->dc_tid);

	} else if (redo_ptr->type == REDO_SELECT) {
	    fprintf(stderr, "error: haven't implemented REDO_SELECT yet\n");
	    exit(1);
	    vista_free(explicit_seg_ptr, redo_ptr, explicit_ptr->dc_tid);
	}
    }
    /*
     * No need to use set_range on redo_tail_ptr, since redo_tail_ptr
     * is recovered manually.
     */
    vista_set_range(dc_seg_ptr, &explicit_ptr->redo_tail_ptr,
		    sizeof(explicit_ptr->redo_tail_ptr),
		    explicit_ptr->dc_tid);
    explicit_ptr->redo_tail_ptr = &explicit_ptr->redo_head;
    explicit_ptr->redo_lastflush_ptr = &explicit_ptr->redo_head;

    vista_set_range(dc_seg_ptr, &(explicit_ptr->redo_head.next_ptr),
		    sizeof(explicit_ptr->redo_head.next_ptr),
		    explicit_ptr->dc_tid);
    explicit_ptr->redo_head.next_ptr = NULL;

    /*
     * This is where dc_recover()'s longjmp returns to.
     */
    if ( (recovering = setjmp(explicit_ptr->context)) == 0) {
	vista_end_transaction(data_seg_ptr, explicit_ptr->data_tid);
#ifdef TWO_PHASE_COMMIT
	/*
	 * Use commit_tentative so 2PC can undo the transaction if needed.
	 */
	vista_commit_tentative(dc_seg_ptr, explicit_ptr->dc_tid);
#else
#ifdef PROG_RETRY
        /* Commit checkpoints if needed */
        chkpttail = explicit_ptr->tail;
        chkpttail = (chkpttail + 1) % VISTA_MAX_TRANS;
        nchk = chkpttail - explicit_ptr ->head ;
        if (nchk < 0)
           nchk += VISTA_MAX_TRANS;
        if ( nchk > MAX_SAVED_CHKPTS )
         {
           /* Commit some checkpoints now and take things of the queue*/
           chkpthead = explicit_ptr->head;
           explicit_ptr->head = (explicit_ptr->head + 1) % VISTA_MAX_TRANS;
           vista_end_transaction(dc_seg_ptr,
                                 explicit_ptr->checkpoint_queue[chkpthead]);
           /* pgr_todo should throw away undo logs here */
           undo_free_ptr = explicit_ptr->undo_records[chkpthead];
           explicit_ptr->undo_records[chkpthead] = (undo_node *)NULL;
         }
        else
         {
           undo_free_ptr = NULL;
         }

        /*  Before commiting update queue of checkpoint ids */

        vista_set_range(dc_seg_ptr,
                     &(explicit_ptr->chkpt_id_queue[explicit_ptr->tail]),
                     sizeof(explicit_ptr->chkpt_id_queue[explicit_ptr->tail]),
                     explicit_ptr->dc_tid);
        explicit_ptr->chkpt_id_queue[explicit_ptr->tail] = explicit_ptr->chkpt_i
d;

        vista_set_range(dc_seg_ptr,
                     &(explicit_ptr->checkpoint_queue[explicit_ptr->tail]),
                     sizeof(explicit_ptr->chkpt_id_queue[explicit_ptr->tail]),
                     explicit_ptr->dc_tid);
        explicit_ptr->chkpt_id_queue[explicit_ptr->tail] = explicit_ptr->chkpt_i
d;

        vista_set_range(dc_seg_ptr,
                     &(explicit_ptr->checkpoint_queue[explicit_ptr->tail]),
                     sizeof(explicit_ptr->checkpoint_queue[explicit_ptr->tail]),
                     explicit_ptr->dc_tid);
        explicit_ptr->checkpoint_queue[explicit_ptr->tail] = explicit_ptr->dc_ti
d;

        vista_set_range(dc_seg_ptr,
                     &(explicit_ptr->undo_records[explicit_ptr->tail]),
                     sizeof(explicit_ptr->undo_records[explicit_ptr->tail]),
                     explicit_ptr->dc_tid);

        explicit_ptr->undo_records[explicit_ptr->tail] = explicit_ptr->undo_head
;

        vista_set_range(dc_seg_ptr, &(explicit_ptr->undo_head),
                    sizeof(explicit_ptr->undo_head), explicit_ptr->dc_tid);
        explicit_ptr->undo_head = NULL;

        vista_set_range(dc_seg_ptr,&(explicit_ptr->tail),
                        sizeof(explicit_ptr->tail), explicit_ptr->dc_tid);
        explicit_ptr->tail = (explicit_ptr->tail + 1) % VISTA_MAX_TRANS;

        vista_commit_tentative(dc_seg_ptr, explicit_ptr->dc_tid);

        /* $$$$ pg_retry code block end pgr_commit */
#else
	vista_end_transaction(dc_seg_ptr, explicit_ptr->dc_tid);
#endif /* PROG_RETRY */

#endif /* TWO_PHASE_COMMIT */

    } else {
	/*
	 * Do miscellaneous tasks for recovery.  These really should be in
	 * dc_recover, but they have to be done after the longjmp.  E.g.
	 * longjmp re-initializes the floating point state, so I have to
	 * restore it back to the way it was at the checkpoint.
	 *
	 * Warning: don't put any code (esp. code that might cause a prot
	 * fault) before calling set_signal_stack.  longjmp apparently messes
	 * up the sigaltstack info.
	 */
	set_signal_stack(explicit_ptr->sig_stack,
	    SP(explicit_ptr->context) >= explicit_ptr->sig_stack &&
	    SP(explicit_ptr->context)<explicit_ptr->sig_stack + MAX_STACK_SIZE);
	float_restore();
    }

    /* start the next transaction */
    explicit_ptr->dc_tid = vista_begin_transaction(dc_seg_ptr, VISTA_IMPLICIT);

#ifdef PROG_RETRY
     vista_set_range(dc_seg_ptr,&(explicit_ptr->chkpt_id),
                      sizeof(explicit_ptr->chkpt_id), explicit_ptr->dc_tid);
     explicit_ptr->chkpt_id++;
#endif /* PROG_RETRY */

    /*
     * Protect the data segment (i.e. data segment starts getting logged here).
     * Any vista calls after this point should be preceded by prefaulting the
     * stack, so that they won't cause faults to vista_sigprot.
     *
     * vista_begin_transaction will cause a prot fault (and I can't prefault it,
     * since the mprotect occurs within vista_begin_transaction).  The fault
     * doesn't lead to concurrency conflicts, because vista_begin_transaction
     * is manipulating the data_seg, and the fault manipulates the dc_seg
     * (which is where things get logged).
     */
    explicit_ptr->data_tid =
	vista_begin_transaction(data_seg_ptr, VISTA_IMPLICIT);

    /*
     * Are we on the signal stack or the normal stack?
     * I could have used sigaltstack and checked if the process was SS_ONSTACK,
     * but checking the value of the stack pointer should be faster (we already
     * have the stack pointer value).
     */
    if (SP(explicit_ptr->context) >= explicit_ptr->sig_stack &&
        SP(explicit_ptr->context) < explicit_ptr->sig_stack + 
	MAX_STACK_SIZE) {

	/* fprintf(stderr, "\ton sigstack\n"); */

	/* must save part of the signal stack */
	exp_start = SP(explicit_ptr->context);

	/* use the normal stack pointer that we saved in dc_sighandler */
	sp = explicit_ptr->sp;

    } else {
	/* Process not handling a signal, so nothing is on signal stack */
	exp_start = explicit_ptr->sig_stack + MAX_STACK_SIZE;

	/* we're on the normal stack */
	sp = SP(explicit_ptr->context);
    }

    if (sp < dc_stack || sp >= dc_stack + MAX_STACK_SIZE) {
	fprintf(stderr, "error: sp not in dc_stack\n");
	exit(1);
    }

    /*
     * Unprotect the empty part of the stack.  We do the mprotect directly
     * (rather than worrying about the application protection bits and using
     * merge_protect), because the application doesn't change protection
     * for the stack.
     * Adding (VISTA_PAGE - 1) then truncating is needed because dc_stack
     * is not page aligned, and we don't want to unprotect memory
     * that's not part of the stack.
     * We don't call mprotect, since that would get intercepted by Vista
     * and interpreted as an application protection change.  We want to
     * manipulate Vista's protection.
     *
     * ??? I really should update the vista_protect array.  It's not
     * a problem to have it be wrong, though, since the application doesn't
     * manipulate the protections on the stack.
     */

    syscall(SYS_mprotect, VISTA_PTR_TO_PAGE(dc_stack + VISTA_PAGE - 1),
	( (sp - (char *) VISTA_PTR_TO_PAGE(dc_stack + VISTA_PAGE - 1) )
	/ VISTA_PAGE) * VISTA_PAGE,
	PROT_READ|PROT_WRITE);

    /* 
     * Pre-fault, so vista_set_range doesn't cause a fault.  This probably
     * isn't needed, since we've unprotected the empty part of the stack.
     */
    prefault();

    /*
     * Log the explicitly logged data that gets changed during every
     * transaction, plus the used part of the signal stack.
     */
    vista_set_range(dc_seg_ptr, exp_start,
		    (char *) &(explicit_ptr->unrecov_data) - exp_start,
		    explicit_ptr->dc_tid);

#ifdef TWO_PHASE_COMMIT
    /*
     * We've now started the next transaction, and we're prepared to
     * commit or abort the last transaction.
     */

    /*
     * Check with the 2PC coordinator to see if the vote is commit or abort.
     */
    got_vote = 0;
    while (!got_vote) {
	/* send prepare (id is explicit_ptr->interval_2pc) */
	message.type = PREPARED;
	message.id = explicit_ptr->interval_2pc;
	message.aborted = explicit_ptr->initiate_abort;

	if (syscall(SYS_sendto, socket_2pc, &message, sizeof(message), 0,
		(struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0) {
	    perror("sendto");
	    exit(1);
	}

	FD_ZERO(&fdset);
	FD_SET(socket_2pc, &fdset);

	/*
	 * Loop until we get a vote, or we time out.  Without this loop,
	 * we'd send a message to the 2PC coordinator each time a message
	 * doesn't match what we're looking for, then the coordinator would
	 * send a message back to us, so the input queue could never drain.
	 */
	while (!got_vote &&
		syscall(SYS_select, OPEN_MAX, &fdset, NULL, NULL, &timeout)
		    != 0) {

	    if (syscall(SYS_recvfrom, socket_2pc, &message, sizeof(message),
		    0, NULL, NULL) == sizeof(message)) {
		if ((message.type == COMMIT || message.type == ABORT) &&
			message.id == explicit_ptr->interval_2pc) {
		    got_vote = 1;
		}
	    }
	}
    }

    if (message.type == ABORT) {
	/*
	 * Server replied "abort".
	 */
	fprintf(stderr, "** 2PC server replied abort\n");

	explicit_ptr->initiate_abort = 0;

	/*
	 * Flush all user sockets.
	 */
	fprintf(stderr, "sleeping 5 seconds\n");
	timeout.tv_sec = 5;
	syscall(SYS_select, 0, NULL, NULL, NULL, &timeout);

	fprintf(stderr, "flushing all user sockets\n");
	timeout.tv_sec = timeout.tv_usec = 0;
	for (fd=USER_FD; fd<OPEN_MAX; fd++) {
	    if (explicit_ptr->fds[fd].status == FD_SOCKET) {
		fprintf(stderr, "flushing fd %d\n", fd);
		FD_ZERO(&fdset);
		FD_SET(fd, &fdset);
		while (
		    syscall(SYS_select, OPEN_MAX, &fdset, NULL, NULL, &timeout)
			> 0) {
		    if (syscall(SYS_recvfrom, fd, &buf, sizeof(buf), 0,
				NULL, NULL) < 0) {
			perror("recvfrom");
			exit(1);
		    }
		}
	    }
	}
	/*
	 * Abort the current transaction (which was just started).
	 */
	dc_abort_internal();
	fprintf(stderr, "done aborting prepare transaction\n");
	/*
	 * Abort the last transaction, which had been tentatively committed.
	 */
	fprintf(stderr, "abort last transaction (tid %d)\n", last_tid);
	vista_abort_transaction(dc_seg_ptr, last_tid);

	/*
	 * dc_recover will finish recovery.  E.g. apply DC's undo log,
	 * re-open files and sockets, restore file positions and file
	 * descriptor state.  dc_recover doesn't return (it longjmp's back
	 * into dc_checkpoint).
	 */
	dc_recover(NULL, 0);
    }

    vista_set_range(dc_seg_ptr, &(explicit_ptr->interval_2pc),
		sizeof(explicit_ptr->interval_2pc), explicit_ptr->dc_tid);
    explicit_ptr->interval_2pc = message.new_id;

    /*
     * Free undo log for last transaction.  Don't free undo log for last
     * transaction if you're recovering (to avoid freeing it twice).
     */
    if (!recovering) {
	vista_end_transaction(dc_seg_ptr, last_tid);
    }
#endif

    /*
     * Free all undo records from the last transaction.  It would have been
     * nice to do this while setting undo_head to NULL (atomically with the
     * end of the last transaction), but vista_free can't handle the fact
     * that the data was malloc'ed in explicit_seg, while the vista_free
     * operation takes place in the dc_seg transaction.
     * Don't do this if you're recovering (don't want to free the list twice)
     */
    if (!recovering) {
	while (undo_free_ptr != NULL) {
	    if (undo_free_ptr->type == UNDO_WRITE) {
		vista_free(explicit_seg_ptr, undo_free_ptr->u.write.data,
			VISTA_NO_TRANS);
	    } else {
		fprintf(stderr, "error: undo_free_ptr->type = %d\n",
			undo_free_ptr->type);
		exit(1);
	    }
	    tmp_undo_ptr = undo_free_ptr->next_ptr;
	    vista_free(explicit_seg_ptr, undo_free_ptr, VISTA_NO_TRANS);
	    undo_free_ptr = tmp_undo_ptr;
	}
    }

    /*
     * If I'm tracking non-deterministics, clear the flag.  Otherwise
     * the flag should always be set (since I would be assuming everything is
     * a non-deterministic event).
     */
#ifdef TRACK_ND
    explicit_ptr->seen_nd = 0;
#else
    explicit_ptr->seen_nd = 1;
#endif

#ifdef TWO_PHASE_COMMIT
    /*
     * If I'm going to initiate an abort, do it ASAP (to avoid doing work
     * that's going to be aborted later anyway).
     */
    if (explicit_ptr->initiate_abort) {
	fprintf(stderr, "initiating abort protocol\n");
	dc_checkpoint(); /* this will get a vote of abort */
    }
#endif

    /* restore signal mask */
    sys_sigprocmask(SIG_SETMASK, &(explicit_ptr->sigmask), NULL);

    /* fprintf(stderr, "end of dc_checkpoint\n"); */

}

/*
 * Recover the process.  command != NULL means this is a new process.
 * command == NULL means this was called from dc_abort, so we shouldn't remap
 * the vista segments.
 */
static void dc_recover(char *command, int abort)
{
    char name[PATH_MAX];
    int i;
    unrecov *unrecov_tmp_ptr;
    sigset_t set;
    int fd;
    undo_node *undo_ptr;
    redo_node *redo_ptr;
    int redo_count;
#ifdef PROG_RETRY
    int j,num_chkpts_available,tail_ptr;
#endif /* PROG_RETRY */

    /*
     * printf might not work (write will) until after restoration is done,
     * because the process crash might have called exit(), which cleans
     * up open I/O streams.
     */

    /*
     * Turn off signals.  Signals will be unmasked after dc_recover returns
     * to dc_checkpoint.  Must use sys_sigprocmask instead of sigprocmask,
     * because my trapped version of sigprocmask unblocks SIG2PC.
     */
    sigfillset(&set);
    sigdelset(&set, VISTA_SIGPROT);
    sys_sigprocmask(SIG_BLOCK, &set, NULL);

    if (command != NULL) {
	/*
	 * explicit needs to be done first, because mapping the data segment
	 * will point unrecov_ptr at an area in the explicit segment, and
	 * unrecov_ptr->lock will be referenced by vista_unlock() at the end
	 * of vista_map() for the data segment.
	 */
	sprintf(name, "%s.%s.explicit", DC_FILE, command);
	if (!(explicit_seg_ptr = vista_map(name, (char *) NULL) ) ) {
	    vista_perror("vista_map of explicit");
	    exit(1);
	}

	/*
	 * Set our copy of Vista's unrecovered data to the values freshly
	 * initialized by vista_init, then point Vista's unrecov_ptr at our copy
	 * (in the explicit segment).
	 * Compute unrecov_ptr manually (through Vista structures) because
	 * explicit_ptr won't be set until data segment is mapped.
	 */
	unrecov_tmp_ptr =
	    &( ((explicit_type *) explicit_seg_ptr->h->base)->unrecov_data);
	memcpy(unrecov_tmp_ptr, unrecov_ptr, sizeof(unrecov));
	unrecov_ptr = unrecov_tmp_ptr;

	/*
	 * vista_map of data_seg overwrites data segment with the data at the
	 * time of crash (including vista's global variables).
	 * I don't need to set data_seg_ptr, explicit_seg_ptr, or dc_seg_ptr,
	 * because these will be restored along with the other global variables.
	 * explicit_seg_ptr is set above only because I need it before mapping
	 * the data segment.
	 */
	sprintf(name, "%s.%s.data", DC_FILE, command);
	if (!vista_map(name, DATA_START)) {
	    vista_perror("vista_map of data");
	    exit(1);
	}

	/*
	 * vista_map of dc segment must be done last.  The other segments must
	 * already be mapped when the undo records for all segments (stored in
	 * dc segment) get processed.
	 */
	sprintf(name, "%s.%s.dc", DC_FILE, command);
	if (!vista_map(name, NULL)) {
	    vista_perror("vista_map of dc");
	    exit(1);
	}

	fprintf(stderr, "done mapping Vista segments\n");
    }

    /*
     * Process DC's undo log
     */ 
    fprintf(stderr, "processing DC undo log\n");
    for (undo_ptr = explicit_ptr->undo_head; undo_ptr != NULL; ) {
        if (undo_ptr->type == UNDO_WRITE) {
            fprintf(stderr, "undoing write to %s, position %d, datasize %d, file
size %d\n",
                undo_ptr->u.write.path, (int) undo_ptr->u.write.position,
                undo_ptr->u.write.datasize, undo_ptr->u.write.filesize);  
            fd = syscall(SYS_open, undo_ptr->u.write.path, O_RDWR);
            if (fd < 0) {  
                fprintf(stderr, "error: open failed during undo processing\n");
                exit(1);
            }
            sys_lseek(fd, undo_ptr->u.write.position, SEEK_SET);
            syscall(SYS_write, fd, undo_ptr->u.write.data,
                    undo_ptr->u.write.datasize);
            sys_ftruncate(fd, undo_ptr->u.write.filesize);
            syscall(SYS_close, fd);
     
            vista_free(explicit_seg_ptr, undo_ptr->u.write.data,
                    VISTA_NO_TRANS);
        } else {
            fprintf(stderr, "error: undo_ptr->type = %d\n", undo_ptr->type);
            exit(1);
        }
        /*
         * Remove undo node from linked list BEFORE freeing it (lest we
         * crash and leave a half-baked undo node on the list)
         */
        explicit_ptr->undo_head = undo_ptr->next_ptr;
        vista_free(explicit_seg_ptr, undo_ptr, VISTA_NO_TRANS);
        undo_ptr = explicit_ptr->undo_head;
    }

    /*
     * Recover DC's redo log.
     */
    explicit_ptr->redo_tail_ptr = &explicit_ptr->redo_head;

    /*
     * Strip out invalid records.
     */
    fprintf(stderr, "eliminating invalid records from redo log\n");
    for (redo_ptr = &explicit_ptr->redo_head; redo_ptr->next_ptr != NULL; ) {
	if (redo_ptr->next_ptr->type == REDO_INVALID) {
	    redo_ptr->next_ptr = redo_ptr->next_ptr->next_ptr;
	} else {
	    redo_ptr = redo_ptr->next_ptr;
	}
    }

    redo_count = 0;
    for (redo_ptr = explicit_ptr->redo_head.next_ptr; redo_ptr != NULL;
	    redo_ptr = redo_ptr->next_ptr, redo_count++) {
	if (redo_ptr->type == REDO_INVALID) {
	    fprintf(stderr, "still have invalid record on redo log\n");
	    exit(1);
	}
    }
    fprintf(stderr, "redo log has %d records\n", redo_count);

#ifdef PROG_RETRY
    if (command != NULL) {
        fprintf(stderr,"Performing Progressive Rollback\n");

     
        num_chkpts_available = explicit_ptr->tail - explicit_ptr->head ;
        if (num_chkpts_available < 0) {
           num_chkpts_available += VISTA_MAX_TRANS;
	}
        if (explicit_ptr->chkpt_id_queue[explicit_ptr->head] == 0) {
           num_chkpts_available -= 1;
	}
        if (explicit_ptr->chkpt_id > explicit_ptr->lastcrash) {
             /*  Crash at a New Point */
           explicit_ptr->lastcrash = explicit_ptr->chkpt_id;
           explicit_ptr->lastretry = ROLL_BACK_INTERVAL;
        } else {
             /*  Need to roll back more */
           explicit_ptr->lastretry += ROLL_BACK_INTERVAL;
        }
        /* Do the rollbacks here  */
        if (explicit_ptr->lastretry > num_chkpts_available) {
           fprintf(stderr,"Cannot rollback more than number of chkpts present:
                    Rollback interval = %d, num of chkpts = %d\n",
                    explicit_ptr->lastretry,  num_chkpts_available);
           explicit_ptr->lastretry = num_chkpts_available;
        }
	for (j=0; j<explicit_ptr->lastretry; j++) {
	    undo_node *undo_ptr;
	    tail_ptr = explicit_ptr->tail-1;
	    if (tail_ptr < 0) {
		tail_ptr += VISTA_MAX_TRANS;
	    }
	    fprintf(stderr,"Rolling back checkpoint %d\n",
		    explicit_ptr->chkpt_id_queue[tail_ptr]);

	    /* pgr_todo start apply the undo logs */
	    for (undo_ptr = explicit_ptr->undo_records[tail_ptr];
		    undo_ptr != NULL;) {
		if (undo_ptr->chkptid !=
			explicit_ptr->chkpt_id_queue[tail_ptr]) {
		    fprintf(stderr,"Checkpoint ids donot match undoid = %d chkid = %d\n",
			    undo_ptr->chkptid,
			    explicit_ptr->chkpt_id_queue[tail_ptr]);
		    exit(1);
		}

		if (undo_ptr->type == UNDO_WRITE) {
		    fprintf(stderr, "undoing write to %s, position %d, datasize %d, file size %d\n",
			    undo_ptr->u.write.path,
			    (int) undo_ptr->u.write.position,
			    undo_ptr->u.write.datasize,
			    undo_ptr->u.write.filesize);
		    fd = syscall(SYS_open, undo_ptr->u.write.path, O_RDWR);
		    if (fd < 0) {
			fprintf(stderr, "error: open failed during undo processing\n");
			exit(1);
		    }

		    sys_lseek(fd, undo_ptr->u.write.position, SEEK_SET);
		    syscall(SYS_write, fd, undo_ptr->u.write.data,
			    undo_ptr->u.write.datasize);
		    sys_ftruncate(fd, undo_ptr->u.write.filesize);
		    syscall(SYS_close, fd);
		    vista_free(explicit_seg_ptr, undo_ptr->u.write.data,
			    VISTA_NO_TRANS);
		} else {
		    exit(1);
		}
		explicit_ptr->undo_records[tail_ptr] = undo_ptr->next_ptr;
		vista_free(explicit_seg_ptr, undo_ptr, VISTA_NO_TRANS);
		undo_ptr = explicit_ptr->undo_records[tail_ptr];
	    }
	    vista_abort_transaction(dc_seg_ptr,
		    explicit_ptr->checkpoint_queue[tail_ptr]);
	}
    }
     
    explicit_ptr->undo_head = (undo_node *)NULL;
#endif /* PROG_RETRY */

    /*
     * Must reset dc_active to 0, since vista_map of data segment will have
     * recovered it back to 1.
     */
    explicit_ptr->dc_active = 0;

    /*
     * Tell Vista to put all implicitly logged data into dc_seg
     */
    vista_set_log(dc_seg_ptr);

    /*
     * set_signal_stack must be called here AND after longjmp returns to
     * dc_checkpoint, even if we're not starting a new process.
     * Apparently, something during recovery in fiddling with different stacks
     * requires this.  Otherwise, you can't mprotect the stack, because the
     * signals will try to run on the normal (write-protected) stack.
     * Must also re-set it after the longjmp; apparently the longjmp messes up
     * the sigaltstack info.
     */
    set_signal_stack(explicit_ptr->sig_stack, 0);

    /*
     * Recover signal handlers.  I think the first signal is 1 (not 0).
     *
     * Note that this doesn't recover signal handlers back to SIG_DFL
     * (this is only relevant for processes that have called
     * dc_abort/dc_abort_internal, since other processes are starting as
     * a fresh process).  E.g. (1) checkpoint (signal handler is SIG_DFL),
     * (2) set signal handler to X, (3) dc_abort.  Process will recover back
     * to (1) (with dc_siglist = SIG_DFL), but I won't do the SYS_sigaction
     * to recover the handler back to SIG_DFL.  Instead, I'll depend on
     * dc_sighandler2 to drop the signal if handler is SIG_DFL.  I do thi
     * because the process might get a spurious signal (where the signal
     * came before the crash), and I don't want it to terminate the process.
     * This is essentially implementing SIG_DFL myself (though only for
     * signals that once had a handler).
     */
    for (i=1; i<NSIG; i++) {
	if (dc_siglist[i].sa_handler != SIG_DFL) {
	    fprintf(stderr, "recovering handler for signal %d\n", i);
	    sigaction(i, dc_siglist+i, NULL);
	}
    }

    /*
     * Re-open files and reset file position.  All entries in this table
     * are for successfully opened files.
     * Don't re-open or reposition vista files; don't reposition non-regular
     * files.
     */
    for (i=0; i<OPEN_MAX; i++) {
	if (explicit_ptr->fds[i].status == FD_FILE ||
	    explicit_ptr->fds[i].status == FD_SPECIAL) {
	    /*
	     * These open calls all completed and succeeded (otherwise they
	     * wouldn't be in the fds table).  So don't recreate or
	     * retruncate them.
	     * May re-position but NOT re-open stdin.
	     */
	    if (i != STDIN_FILENO) {
		fprintf(stderr, "re-opening file fd %d (%s)\n", i,
			explicit_ptr->fds[i].u.file.path);
		fd = syscall(SYS_open, explicit_ptr->fds[i].u.file.path,
		    explicit_ptr->fds[i].u.file.flags &
			(~(O_CREAT|O_EXCL|O_TRUNC)));
		if (fd != i) {
		    fprintf(stderr, "error: re-opened fd %d, but got %d\n",
					i, fd);
		    if (fd < 0) {
			perror("open");
		    }
		    exit(1);
		}
	    }
	    if (explicit_ptr->fds[i].status == FD_FILE) {
		fprintf(stderr, "repositioning fd %d to %d\n",
				    i, explicit_ptr->fds[i].u.file.position);
		sys_lseek(i, explicit_ptr->fds[i].u.file.position, SEEK_SET);
	    }
	} else if (explicit_ptr->fds[i].status == FD_SOCKET) {
	    fprintf(stderr, "re-socketing fd %d\n", i);
	    fd = syscall(SYS_socket, explicit_ptr->fds[i].u.sock.domain,
		    explicit_ptr->fds[i].u.sock.type,
		    explicit_ptr->fds[i].u.sock.protocol);
	    if (fd != i) {
		fprintf(stderr, "error: re-socketed fd %d, but got %d\n",
				    i, fd);
		if (fd < 0) {
		    perror("socket");
		}
		exit(1);
	    }

	    if (explicit_ptr->fds[i].u.sock.bound) {
		fprintf(stderr, "rebinding fd %d\n", i);
		if (syscall(SYS_bind, i,
			(struct sockaddr *) &explicit_ptr->fds[i].u.sock.addr,
			sizeof(explicit_ptr->fds[i].u.sock.addr))) {
		    perror("bind");
		    exit(1);
		}
	    }
	    if (explicit_ptr->fds[i].u.sock.sigio) {
		fprintf(stderr, "resetting fd %d for sigio to current process\n", i);
		if (syscall(SYS_fcntl, i, F_SETOWN, syscall(SYS_getpid)) < 0) {
		    perror("fcntl (F_SETOWN)");
		    exit(1);
		}
		if (syscall(SYS_fcntl, i, F_SETFL, O_ASYNC) < 0) {
		    perror("fcntl (F_SETFL)");
		    exit(1);
		}
	    }
	    if (explicit_ptr->fds[i].u.sock.nonblock) {
		fprintf(stderr, "resetting fd %d to be nonblocking\n", i);
		if (syscall(SYS_fcntl, i, F_SETFL, O_NONBLOCK) < 0) {
		    perror("fcntl (F_SETFL)");
		    exit(1);
		}
	    }
	    if (explicit_ptr->fds[i].u.sock.rcvbuf > 0) {
		fprintf(stderr, "resetting SO_RCVBUF for fd %d to %d\n",
		    i, explicit_ptr->fds[i].u.sock.rcvbuf);
		if (syscall(SYS_setsockopt, i, SOL_SOCKET, SO_RCVBUF,
		    &(explicit_ptr->fds[i].u.sock.rcvbuf),
		    sizeof(explicit_ptr->fds[i].u.sock.rcvbuf))) {
		    perror("setsockopt");
		    exit(1);
		}
	    }
	    if (explicit_ptr->fds[i].u.sock.sndbuf > 0) {
		fprintf(stderr, "resetting SO_SNDBUF for fd %d to %d\n",
		    i, explicit_ptr->fds[i].u.sock.sndbuf);
		if (syscall(SYS_setsockopt, i, SOL_SOCKET, SO_SNDBUF,
		    &(explicit_ptr->fds[i].u.sock.sndbuf),
		    sizeof(explicit_ptr->fds[i].u.sock.sndbuf))) {
		    perror("setsockopt");
		    exit(1);
		}
	    }
	    if (explicit_ptr->fds[i].u.sock.broadcast) {
		fprintf(stderr, "resetting SO_BROADCAST for fd %d to %d\n",
		    i, explicit_ptr->fds[i].u.sock.broadcast);
		if (syscall(SYS_setsockopt, i, SOL_SOCKET, SO_BROADCAST,
		    &(explicit_ptr->fds[i].u.sock.broadcast),
		    sizeof(explicit_ptr->fds[i].u.sock.broadcast))) {
		    perror("setsockopt");
		    exit(1);
		}
	    }
	} else if (i >= USER_FD) {
	    /*
	     * make sure future opens get the right fd (open uses the
	     * lowest-numbered empty fd).
	     */
	    fd = syscall(SYS_open, "/dev/null", O_RDWR);
	}
    }

    /*
     * Close the dummy file descriptors.
     */
    for (i=USER_FD; i<OPEN_MAX; i++) {
	if (explicit_ptr->fds[i].status == FD_CLOSED) {
	    if (syscall(SYS_close, i) < 0) {
		perror("close");
		exit(1);
	    }
	}
    }

    /*
     * Register the new pid in the pid table file.
     */
    set_current_pid();

    /*
     * Restore timer value.
     */
    setitimer(ITIMER_REAL, &timer, NULL);

    /*
     * Restoring the floating point state must be done after longjmp returns to
     * dc_checkpoint.
     */

    /*
     * Call any user-specific recovery (e.g. re-initialize curses).
     */
    if (user_recov_ptr != NULL) {
	(*user_recov_ptr)();
    }

    explicit_ptr->dc_active = 1;

    /*
     * Figure out if I need to initiate an abort.  Abort if the process is
     * restarting after a crash, or the user initiated a dc_abort.
     */
#ifdef TWO_PHASE_COMMIT
    if (abort && explicit_ptr->sent_message) {
	explicit_ptr->initiate_abort = 1;
    }
#endif
    /* fprintf(stderr, "dc_recover about to call longjmp\n"); */

    longjmp(explicit_ptr->context, 1);
}

/*
 * Revert back to the last checkpoint.
 */
void dc_abort()
{
    fprintf(stderr, "dc_abort\n");

    dc_abort_internal();
    /*
     * dc_recover will finish recovery.  E.g. apply DC's undo log, re-open
     * files and sockets, restore file positions and file descriptor state.
     */
    dc_recover(NULL, 1);

    /* this is never reached */
}

/*
 * Revert back to the last checkpoint.  Most of the work is done in
 * dc_recover.  dc_abort_internal just prepares to call dc_recover by closing
 * all user files and aborting the current transaction.
 */
static void dc_abort_internal()
{
    jmp_buf context;
    long stack_size;
    int i;
    char *tmp_stack;

    tmp_stack = explicit_ptr->tmp_stack;

    /*
     * Close all user files (dc_recover will reopen the ones that were
     * open at the time of the last checkpoint).
     */
    for (i=USER_FD; i<OPEN_MAX; i++) {
	if (explicit_ptr->fds[i].status != FD_CLOSED) {
	    syscall(SYS_close, i);
	}
    }
    /*
     * Switch to a temporary stack so I can revert back to a prior stack
     * state without worrying about affecting dc_abort_internal().  I used
     * to use the original system stack as the temporary stack, but this
     * caused problems (perhaps the original system stack had some other
     * state on it, e.g. written by sigaltstack?).  Instead, I now use
     * a stack in explicit segment just for this purpose.
     *
     * current stack goes from [$sp, $sp + stack_size), where
     * 	    stack_size = dc_stack + MAX_STACK_SIZE - $sp
     * temporary goes from
     *   [tmp_stack + MAX_STACK_SIZE - stack_size, tmp_stack + MAX_STACK_SIZE)
     */
    if (!setjmp(context)) {
	/*
	 * Initialize temporary stack from current stack.
	 * Stack size calculation depends on which stack we're on (dc_stack
	 * or sig_stack).
	 */
	if (SP(context) >= dc_stack && SP(context) < dc_stack+MAX_STACK_SIZE) {
	    stack_size = dc_stack + MAX_STACK_SIZE - SP(context);
	} else {
	    stack_size = explicit_ptr->sig_stack + MAX_STACK_SIZE - SP(context);
	}
	memcpy(tmp_stack + MAX_STACK_SIZE - stack_size, SP(context),
		stack_size);

	/*
	 * Adjust frame and stack pointers.  Make sure to adjust FP first,
	 * since the equations use SP.
	 */
	FP(context) += tmp_stack + MAX_STACK_SIZE - (stack_size + SP(context));
	SP(context) += tmp_stack + MAX_STACK_SIZE - (stack_size + SP(context));

	longjmp(context, 1);

	/* this is never reached */
    }
    /*
     * Change the data in the vista segments back to the last checkpoint.
     */
    vista_abort_transaction(data_seg_ptr, explicit_ptr->data_tid);
    vista_abort_transaction(dc_seg_ptr, explicit_ptr->dc_tid);
}

/*
 * dc_exit is used by the program to really exit, i.e. the program won't
 * be recoverable after a dc_exit.
 */
void dc_exit(int status)
{
    char name[PATH_MAX];
    int i;
    int got_deregistered;
#ifdef TWO_PHASE_COMMIT
    msg message;
    fd_set fdset;
    struct timeval timeout = {1, 0};
#endif

    fprintf(stderr, "dc_exit making the process unrecoverable\n", dc_command);
#ifdef TWO_PHASE_COMMIT
    do {
	/* send register message */
	message.type = DEREGISTER;

	if (syscall(SYS_sendto, socket_2pc, &message, sizeof(message), 0,
		(struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0) {
	    perror("sendto");
	    exit(1);
	}

	FD_ZERO(&fdset);
	FD_SET(socket_2pc, &fdset);

	got_deregistered = 0;
	if (syscall(SYS_select, OPEN_MAX, &fdset, NULL, NULL, &timeout) == 1) {
	    if (syscall(SYS_recvfrom, socket_2pc, &message, sizeof(message),
		    0, NULL, NULL) == sizeof(message)) {
		if (message.type == DEREGISTERED) {
		    got_deregistered = 1;
		    explicit_ptr->interval_2pc = message.new_id;
		}
	    }
	}
    } while (!got_deregistered);
#endif

    for (i=USER_FD; i<OPEN_MAX; i++) {
	if (explicit_ptr->fds[i].status != FD_CLOSED) {
	    if (syscall(SYS_close, i) < 0) {
		perror("close");
		exit(1);
	    }
	}
    }

    /*
     * Create the file dc.<command>.done, which tells DC that the process
     * really exited.  The next time DC is run, it will delete all the
     * DC files.
     * I ran into problems trying to unlink DC files here.  Unlinking dc.data
     * wipes out the data segment (including the stack) while the process is
     * still running.  Also, unlinking the other files causes I/O when the
     * process exits (exit() -> exit1() -> vm_map_remove() -> vm_map_delete()
     * -> vm_map_entry_delete() -> vm_object_deallocate() ->
     * vm_object_terminate() -> vm_object_page_clean() -> ffs_fsync()).
     */
    sprintf(name, "%s.%s.done", DC_FILE, dc_command);
    syscall(SYS_open, name, O_CREAT|O_TRUNC|O_WRONLY, 0666);
    exit(status);
}

/*
 * DC's 1st signal handler.  Saves the pointer into the normal stack, then calls
 * vista_sigprot.  The pointer into the normal stack is used
 * to unprotect the empty part of the normal stack.  This is only needed if
 * the application's sigprot_handler (or another signal handler) calls
 * dc_checkpoint (which TreadMarks will do a lot by sending a message).
 */
static void dc_sighandler(int sig, int code, struct sigcontext *sc, char* vaddr)
{
    /*
     * only update sp if signal occurred while on the normal stack.
     */
    if (SP(*sc) >= dc_stack && SP(*sc) < dc_stack + MAX_STACK_SIZE) {
	explicit_ptr->sp = SP(*sc);
    }

    if (sig == VISTA_SIGPROT) {
	vista_sigprot(sig, code, sc, vaddr);
    } else {
	/*
	 * Don't allow any signals while replaying the redo log (except
	 * SIGPROT).  Otherwise replay might not follow the pre-crash path.
	 *
	 * ??? I should really mask signals until the last non-abortable
	 * event is executed, but that requires logging the non-abortables
	 * (or logging every few seconds and assuming replay ends a few
	 * seconds after the last log entry).
	 */
	if (explicit_ptr == NULL || explicit_ptr->redo_tail_ptr == NULL ||
		explicit_ptr->redo_tail_ptr->next_ptr == NULL ||
		explicit_ptr->redo_tail_ptr->next_ptr->type == REDO_INVALID) {
	    vista_sigother(sig);
	} else {
	    fprintf(stderr, "in replay, dropping signal %d\n", sig);
	}
    }
}

/*
 * DC's 2nd signal handler.  This is called by vista_sigother when the
 * signal is actually going to be handled.  dc_sighandler2 exists to take
 * checkpoints (since getting a signal is non-deterministic).  After this,
 * dc_sighandler2 then calls the user's signal handler.  Note that the
 * vista_lock will not be held when dc_sighandler2 is called, since
 * vista_sigother only calls it when the vista_lock is free.
 */
static void dc_sighandler2(int sig)
{
    /*
     * Receiving a user's signal is non-deterministic.
     * Don't checkpoint for SIG2PC because the signal handler itself is
     * dc_checkpoint.
     */
    if (sig != SIG2PC) {
	seen_nd("dc_sighandler2");
    }

    if (dc_siglist[sig].sa_handler == SIG_DFL) {
	/*
	 * This only comes up when a process had a signal registered, then
	 * called dc_abort/dc_abort_internal which recovered the process back
	 * to a point before the signal was registered, then had a spurious
	 * signal.  This is somewhat common, because there are often pending
	 * signals that happen during the recovery.  See note in dc_recover
	 * (when recovering signals).
	 */
	fprintf(stderr, "dc_sighandler2: ignoring signal %d\n", sig);
    } else {
	(*dc_siglist[sig].sa_handler)(sig);
    }
}

/*
 * set_signal_stack:
 * Modify sigaction to process signals on sigstack, and use DC's prot
 * handler (which will then call vista_sigprot)
 */
static void set_signal_stack(char *ptr, int onstack)
{
    struct sigaltstack sigstack_desc;
    struct sigaction action;

    /*
     * Tell system to process signals on sigstack
     */
    sigstack_desc.ss_sp = ptr;
    sigstack_desc.ss_size = MAX_STACK_SIZE;
    if (onstack) {
	sigstack_desc.ss_flags = SS_ONSTACK;
    } else {
	sigstack_desc.ss_flags = 0;
    }
    if (sigaltstack(&sigstack_desc, NULL) < 0) {
	fprintf(stderr, "sigaltstack failed\n");
	perror("sigaltstack");
	exit(1);
    }
}

static void prefault()
{
    /*
     * Pre-fault the stack.  This prevents Vista routines from faulting the
     * stack (which would cause them to enter vista_sigprot).
     */
    char buf[4096];

    if (dc_seg_ptr == NULL) {
	fprintf(stderr, "error: tried to use an dc function before dc_init\n");
	while(1);
    }

    buf[0] = buf[4095] = 1;
}

static void delete(char *command, char *suffix)
{
    char name[PATH_MAX];
    sprintf(name, "%s.%s.%s", DC_FILE, command, suffix);
    syscall(SYS_unlink, name);
    sprintf(name, "%s.%s.%s.meta", DC_FILE, command, suffix);
    syscall(SYS_unlink, name);
}

static int file_exists(char *command, char *suffix)
{
    char name[PATH_MAX];
    int fd;

    sprintf(name, "%s.%s.%s", DC_FILE, command, suffix);
    if ( (fd = open(name, O_RDONLY)) >= 0) {
	close(fd);
	return(1);
    } else {
	return(0);
    }
}

static int fresh_process(char *command)
{
    if (file_exists(command, "done") ||
	    !file_exists(command, "explicit") ||
	    !file_exists(command, "explicit.meta") ||
	    !file_exists(command, "dc") ||
	    !file_exists(command, "dc.meta") ||
	    !file_exists(command, "data") ||
	    !file_exists(command, "data.meta")) {

	/*
	 * Clean up
	 */
	delete(command, "dc");
	delete(command, "explicit");
	delete(command, "data");
	delete(command, "done"); /* delete also spuriously deletes done.meta
				    (which doesn't exist), but this is harmless
				    */
	delete(command, "pipe"); /* delete also spuriously deletes pipe.meta
				    (which doesn't exist), but this is harmless
				    */
	return(1);
    } else {
	return(0);
    }
}

static void float_save()
{
    char *fdata;

    vista_set_range(dc_seg_ptr, &(explicit_ptr->floating_point[0]),
		    FLOAT_SIZE * sizeof(explicit_ptr->floating_point[0]),
		    explicit_ptr->dc_tid);

    fdata = explicit_ptr->floating_point;
    __asm__("fsave %0": :"m"(*fdata));

    /*
     * fsave re-initializes the floating-point unit, so use frstor to re-load
     * the state.
     */
    __asm__("frstor %0": :"m"(*fdata));
}

static void float_restore()
{
    char *fdata;

    fdata = explicit_ptr->floating_point;
    __asm__("frstor %0": :"m"(*fdata));
}

/*
 * exit.  Only used for performance measurements (to print final statistics).
 * ???
 */
#include "atexit.h"

void (*__cleanup)();

/*
 * Exit, flushing stdio buffers if necessary.
 */
void
exit(status)
	int status;
{
	register struct atexit *p;
	register int n;
	struct timeval t_end;
	char buf[15][100];
	int buf_num = 0;
	int fd;
	int i;

	for (p = __atexit; p; p = p->next)
		for (n = p->ind; --n >= 0;)
			(*p->fns[n])();

	/*
	 * Prevent any further checkpoints.
	 */
	explicit_ptr->dc_active = 0;

	syscall(SYS_gettimeofday, &t_end, NULL);

	/*
	if (freopen("/tmp/dc.finalStats.tmp", "w", stderr) != stderr) {
	    perror("freopen");
	    _exit(1);
	}
	*/

	/*
	 * Print the type of recovery protocol that was used in this run.
	 */
#ifdef DC_DISK
	sprintf(buf[buf_num++], "DC_DISK\n");
#else
	sprintf(buf[buf_num++], "No DC_DISK\n");
#endif

#ifdef APP_MPROTECT
	sprintf(buf[buf_num++], "APP_MPROTECT\n");
#else
	sprintf(buf[buf_num++], "No APP_MPROTECT\n");
#endif

#if defined(TRACK_NA) && !defined(TRACK_ND) && !defined(REDO_LOG) && !defined(TWO_PHASE_COMMIT)
	sprintf(buf[buf_num++], "CBNA\n");
#elif !defined(TRACK_NA) && defined(TRACK_ND) && !defined(REDO_LOG) && !defined(TWO_PHASE_COMMIT)
	sprintf(buf[buf_num++], "CAND\n");
#elif defined(TRACK_NA) && defined(TRACK_ND) && !defined(REDO_LOG) && !defined(TWO_PHASE_COMMIT)
	sprintf(buf[buf_num++], "CB\n");
#elif !defined(TRACK_NA) && defined(TRACK_ND) && defined(REDO_LOG) && !defined(TWO_PHASE_COMMIT)
	sprintf(buf[buf_num++], "LOG\n");
#elif defined(TRACK_NA) && defined(TRACK_ND) && defined(REDO_LOG) && !defined(TWO_PHASE_COMMIT)
	sprintf(buf[buf_num++], "CBLOG\n");
#elif defined(TRACK_NA) && !defined(TRACK_ND) && !defined(REDO_LOG) && defined(TWO_PHASE_COMMIT)
	sprintf(buf[buf_num++], "TWOPHASE\n");
#elif defined(TRACK_NA) && defined(TRACK_ND) && !defined(REDO_LOG) && defined(TWO_PHASE_COMMIT)
	sprintf(buf[buf_num++], "CBTWOPHASE\n");
#else
#	error "unknown recovery protocol"
#endif

	sprintf(buf[buf_num++], "duration %lf seconds\n",
	    (t_end.tv_sec - t_start.tv_sec) +
	    (t_end.tv_usec - t_start.tv_usec) / 1000000.0);

	sprintf(buf[buf_num++], "numCheckpoints %d\n",
		explicit_ptr->num_checkpoints);

	if (bytes_logged > max_bytes_logged) {
	    max_bytes_logged = bytes_logged;
	}

	/*
	 * Don't count the last interval towards the average.
	 */
	sprintf(buf[buf_num++], "lastIntervalBytesLogged %d\n", bytes_logged);
	total_bytes_logged -= bytes_logged;

	sprintf(buf[buf_num++], "maxBytesLogged %d\n", max_bytes_logged);
	if (explicit_ptr->num_checkpoints > 1) {
	    sprintf(buf[buf_num++], "averageBytesLogged %d\n", 
		(int) (total_bytes_logged / (explicit_ptr->num_checkpoints - 1))
		);
	} else {
	    sprintf(buf[buf_num++], "averageBytesLogged %d\n", 
	    (int) (total_bytes_logged) );
	}

	if (__cleanup) {
	    (*__cleanup)();
	}

	/*
	 * Print final statistics.  Do this in a way that generates minimal
	 * other work for DC (e.g. freopen sometimes faults).
	 */
	fd = syscall(SYS_open, "/tmp/dc.finalStats.tmp", O_WRONLY | O_TRUNC | O_CREAT, 0777);

	for (i=0; i<buf_num; i++) {
	    syscall(SYS_write, fd, buf[i], strlen(buf[i]));
	}

	_exit(status);
}

/*
 * DC sees a non-deterministic event--take the appropriate actions (e.g.
 * checkpoint).
 * This is a public function to allow performance tests to explicitly
 * notify DC if they've done something non-deterministic that DC would
 * normally assume is deterministic.  The only example of this so far is
 * a simulated keyboard that takes its input from a file (used to make
 * performance tests more repeatable).
 */
void seen_nd(char *message)
{
    /* fprintf(stderr, "seen_nd %s\n", message); */
    explicit_ptr->seen_nd = 1;

    /*
     * If we're not tracking non-abortables, we must checkpoint after every
     * non-deterministic.
     */
#ifndef TRACK_NA
    if (explicit_ptr && explicit_ptr->dc_active && explicit_ptr->seen_nd) {
	fprintf(stderr, "seen_nd calling dc_checkpoint (%s)\n", message);
	dc_checkpoint();
    }
#endif
}

/*
 * DC sees a non-abortable event--take the appropriate actions (e.g.
 * checkpoint).  This only does something when DC is tracking non-abortables
 * (otherwise DC is unaware of non-abortables and must assume everything
 * is a non-abortable, e.g. by checkpointing after each non-deterministic
 * and by flushing the redo log after each redo record is created.
 *
 * This is a public function to allow the user to explicitly say they're about
 * to do a non-abortable event.  I haven't thought of a use for this yet,
 * but it might come in handy.
 */
void seen_na(char *message)
{
#ifdef TRACK_NA
    /* fprintf(stderr, "seen_na %s\n", message); */

    /*
     * Flush the redo log to disk (only has an effect for DC_DISK).
     */
    redo_flush();

    if (explicit_ptr && explicit_ptr->dc_active && explicit_ptr->seen_nd) {
	/* fprintf(stderr, "seen_na calling dc_checkpoint (%s)\n", message); */
	dc_checkpoint();
    }
#endif
}

/*
 * Flush the redo log to Vista's redo log on disk.  This function only does
 * something when using DC_DISK.  Vista handles its own log flush of the
 * new data in a transaction (this isn't complete yet; e.g. I don't think Vista
 * flushes the heap operation log).
 */
static void redo_flush()
{
#if defined(DC_DISK) && defined(REDO_LOG)
    redo_node *redo_ptr;

    dc_lock();

    for (redo_ptr = explicit_ptr->redo_lastflush_ptr;
	    redo_ptr->next_ptr != NULL; redo_ptr = redo_ptr->next_ptr) {

	/*
	 * Write out record at redo_ptr.
	 */
	if (redo_ptr->type == REDO_RECEIVE || redo_ptr->type == REDO_READ) {
	    if (redo_ptr->select_buf_ptr[0] != NULL) {
		syscall(SYS_write, dc_seg_ptr->redo_log,
			redo_ptr->select_buf_ptr[0],
			sizeof(*redo_ptr->select_buf_ptr[0]));
		syscall(SYS_write, dc_seg_ptr->redo_log,
			redo_ptr->select_buf_ptr[0]->buf,
			redo_ptr->select_buf_ptr[0]->length);
	    }

	} else if (redo_ptr->type == REDO_SELECT) {
	    fprintf(stderr, "error: haven't implemented REDO_SELECT yet\n");
	    exit(1);
	}
	syscall(SYS_write, dc_seg_ptr->redo_log, redo_ptr,
		sizeof(*redo_ptr));
    }
    explicit_ptr->redo_lastflush_ptr = redo_ptr;

    fsync(dc_seg_ptr->redo_log);

    dc_unlock();
#endif
}
