#include <setjmp.h>
#include <signal.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>

#ifdef DC_INTERNAL

#include "include/vista.h"
#include "include/lock.h"
#define DC_FILE "/usr/tmp/dc"
#define SP(context) ( (char *) (((struct sigcontext *) &(context))->sc_sp))

#endif /* DC_INTERNAL */

#define MAX_STACK_SIZE (1*1024*1024)
#define FLOAT_SIZE 108

extern unsigned long etext;
/* round DATA_START up to nearest page */
#define DATA_START ( (char *) ( ( ( (long)&etext-1) / 4096 + 1) * 4096 ))
#define STACK_START ((char *) 0xefbfe000)

#define FP(context) ( (char *) (((struct sigcontext *) &(context))->sc_fp))

/*
 * dc_init must be a macro and must be called from main.  This is because
 * it changes main's frame pointer to the new stack (in Vista space).
 * If main() called func(), and func() called dc_init, main's frame pointer
 * would still get restored from the value saved on the stack before
 * the stack got switched to Vista space.
 */
#define dc_init(command, check_ptr, recov_ptr, argvp, envpp) {	\
    jmp_buf context;						\
    dc_init_internal(command, check_ptr, recov_ptr, argvp, envpp);	\
    if (!setjmp(context)) {						\
	FP(context) -= STACK_START - (dc_stack_public + MAX_STACK_SIZE );\
	longjmp(context, 1);					\
    }								\
    dc_checkpoint();						\
}

#ifdef DC_INTERNAL

#define MAX_MESSAGE (64 * 1024)	    /* max UDP message is 64 KB */
#define SPECIAL_BLOCK 512	    /* # of bytes to read when saving pending
					read */

typedef struct {
    char *buf;    		    /* message data */
    ssize_t length;		    /* size of the data */
    struct sockaddr_in from;	    /* sender's address */
    int fromlen;		    /* size of sender's address */
} selbuf;

typedef struct {

#define FD_CLOSED 0
#define FD_FILE 1	/* regular files.  Recover their position after
			    a crash.  Writes to these files are undoable. */
#define FD_SPECIAL 2	/* files that are not regular files, e.g. /etc/tty.
			    Reopen them during recovery, but don't reposition
			    them. */
#define FD_SOCKET 3

    int status;
    selbuf *select_buf;		/* space to save the first pending message after
				    a select */
    union {
	struct {
	    char path[PATH_MAX];
	    int flags;
	    int unlink;
	    off_t position;	/* current position */
	    int pipe;		/* is this used for a pipe?  If so, checkpoint
				    before each write (since we know another
				    process is reading the file) */
	} file;
	struct {
	    int domain;
	    int type;
	    int protocol;
	    int bound;			    /* socket is bound */
	    struct sockaddr_in addr;	    /* addr. this socket is bound to */
	    int connected;		    /* is the socket connected? */
	    struct sockaddr_in connaddr;
	    int sigio;			    /* send SIGIO to current process */
	    int nonblock;		    /* socket is nonblocking */
	    int sndbuf;			    /* size of the SO_SNDBUF */
	    int rcvbuf;			    /* size of the SO_RCVBUF */
	    int broadcast;		    /* SO_BROADCAST */
	} sock;
    } u;
} fd_info;

typedef struct redo_struct {
#define REDO_INVALID 0
#define REDO_RECEIVE 1
#define REDO_READ 2
#define REDO_SELECT 3
    int type;
    int fd;				/* fd and buf check to make sure we're
					    replaying the right log record */
    char *buf;
    int return_errno;
    int return_status;
    selbuf *select_buf_ptr[OPEN_MAX];	/* array of pointers.  Each pointer
					    points to a select buffer.  We
					    need an array because a select
					    log record needs to point to
					    several select buffers. */
    struct redo_struct *next_ptr;
} redo_node;

typedef struct undo_struct {

#ifdef PROG_RETRY
    int chkptid;
#endif /* PROG_RETRY */

#define UNDO_WRITE 0		    /* undo a write to a normal file */
    int type;			    /* type of undo record */
    union {
	struct {
	    char *path;		    /* null-terminated path string */
	    off_t position;	    /* offset in file */
	    char *data;		    /* original data in file */
	    size_t datasize;	    /* size of write data */
	    off_t filesize;	    /* size of file before the write */
	} write;
    } u;
    struct undo_struct *next_ptr;
} undo_node;

#ifdef PROG_RETRY

#define MAX_SAVED_CHKPTS 30
#define ROLL_BACK_INTERVAL 5

#endif /* PROG_RETRY */


/*
 * Explicitly logged data.  Data is put here for one of several reasons:
 *  1) Variable (viz. sigmask, context, sp) is written in every transaction,
 *     so it's cheaper to log them explicitly.
 *  2) Variable (viz. unrecov_data, dc_tid, data_tid) does not need to be
 *     logged, because it will be re-initialized during Vista's recovery.
 *  3) sigprot needs to operate on a stack that doesn't cause further sigprot
 *     faults.  Hence sig_stack is explicitly logged.  For most applications
 *     (those that don't use sigprot, or whose sigprot handlers don't call
 *     dc_checkpoint), sig_stack will be empty when dc_checkpoint is called.
 *     Hence sig_stack is never logged for most applications.
 * This structure will be malloc'ed out of explicit_seg.
 */
typedef struct {
    char sig_stack[MAX_STACK_SIZE]; /* placed first in explicit_type so that
					the empty part of the signal stack
					is at the end (not in between the
					used part of the signal stack and
					the other variables */
    sigset_t sigmask;	/* signal mask as the beginning of the current
			    transaction */
    jmp_buf context;	/* registers, PC, SP, etc. at last checkpoint */
    char floating_point[FLOAT_SIZE];
			/* floating point state at last checkpoint */
    char *sp;	/* stack pointer before the sigprot handler switches to
		    sig_stack.  Used to unprotect empty part of normal stack. */

    /*
     * variables below this line are not logged (unless dc does an explicit
     * set_range, as in the files array).  unrecov_data should be
     * the first variable, since it's used by dc_checkpoint as a delimiter.
     */
    unrecov unrecov_data; /* A copy of Vista's unrecoverable data.  We point
			    Vista's unrecov_ptr at our copy in the explicit
			    segment and DON'T log it.  This way vista_map's
			    during DC's recovery doesn't change vista's
			    initialized variables.  Some of these variables
			    really can't be implicitly logged anyway (e.g.
			    lock is written during Vista's sighandler;
			    defer could be written at any time by the sigio
			    handler).  */
    int dc_tid;
    int data_tid;
    int dc_active;	/* enable automatic checkpoints.  Should be disabled
				during recovery */
    fd_info fds[OPEN_MAX];	/* info about which fds the program
				    has opened and closed */
    undo_node *undo_head;	/* linked list of undo records for operations
				    (e.g. file writes) that DC is responsible
				    to undo */
    redo_node redo_head, *redo_tail_ptr;
				/* redo_tail_ptr always points to the last
				    replayed record.  On recovery, it points
				    to redo_head.  redo_head is the first
				    record on the redo log.  It's never
				    replayed (just a placeholder for the head
				    of the redo log). */
    redo_node *redo_lastflush_ptr; /* points to the last redo record that has
				    been flushed to disk (when using DC_DISK) */

    char tmp_stack[MAX_STACK_SIZE]; /* temporary stack used in recovery */
    int num_checkpoints;	/* number of checkpoints taken */
#ifdef PROG_RETRY
    int checkpoint_queue[VISTA_MAX_TRANS]; /* Circular Queue of checkpoints for
                                                progressive retry constrained
                                                by vista transactions*/
    int chkpt_id_queue[VISTA_MAX_TRANS];
    undo_node *undo_records[VISTA_MAX_TRANS];
    int head,tail;
    int chkpt_id;
    int lastretry,lastcrash;
#endif /* PROG_RETRY */

#ifdef TWO_PHASE_COMMIT
    int sent_message;	/* have I sent messages in the current interval
			    (that need to be aborted)?  Used in 2PC */
    int interval_2pc;	/* ID for 2PC interval */
    int initiate_abort;		/* process wants to initiate a 2PC abort */
#endif
    int seen_nd;	/* have I seen a non-deterministic since the last
			    checkpoint? */
    struct timeval last_time;	/* ??? used in performance measurements */

} explicit_type;

#define SIG2PC SIGEMT

/*
 * Private functions
 */
static void dc_recover(char *, int);
static void set_signal_stack(char *, int);
static void prefault();
static void delete(char *, char *);
static void dc_sighandler();
static void dc_sighandler2();
static void dc_abort_internal();
static int sig_norestart();
static int fresh_process(char *);
static void float_save();
static void float_restore();
static void redo_flush();

#endif /* DC_INTERNAL */

/*
 * dc_stack need to be visible to users of dc because dc_init is a macro.
 * Make it a char * so I can assign it a different value in dc_base.c (which
 * is a special DC hacked to have no checkpointing overhead).
 */
extern char *dc_stack_public;

/*
 * Public functions 
 */
void dc_checkpoint();
int dc_touch(char *, int);
void dc_exit(int status);
void seen_nd(char *);
void seen_na(char *);
int dc_select(int nfds, fd_set *readfds, fd_set *writefds, fd_set *exceptfds,
    struct timeval *timeout);

/* dc_init, which is a macro */
void dc_init_internal(char *command, void (*check_ptr)(), void (*recov_ptr)(),
    char ***, char ***);
