/*
 * sockets.c
 *
 *	This module is intended to subvert an application's calls to 
 *	most or all TCP socket syscalls, and to redirect them so that they
 *	use mocktcp instead of the system's tcp. This module should not
 *	be used in applications that use UDP.
 */

#include "mocktcp.h"
#include "namespace.h"
#include <sys/syscall.h>

int socket(int domain, int type, int protocol)
{
	return mt_socket();
}

int connect(int s, const struct sockaddr *name, int namelen)
{
	return mt_connect_sock(s, name);
}

int listen(int s, int backlog)
{
	return 0;
}

int accept(int s, struct sockaddr *addr, int *addrlen)
{
	int	res;

	res = mt_accept(s, (struct sockaddr_in*)addr);
	if (res >= 0) {
		*addrlen = sizeof(struct sockaddr_in);
	}
	return res;
}

int send(int s, const void *msg, size_t len, int flags)
{
	if (flags & MSG_OOB) {
		fprintf(stderr, "warning: mocktcp told to send OOB data\n");
		flags &= (~MSG_OOB);
	}
	if (flags & MSG_DONTROUTE) {
		fprintf(stderr, "warning: mocktcp told to send DONTROUTE\n");
		flags &= (~MSG_DONTROUTE);
	}
	if (flags & MSG_EOR) {
		fprintf(stderr, "warning: mocktcp told to send EOR\n");
		flags &= (~MSG_EOR);
	}
	if (flags & MSG_EOF) {
		fprintf(stderr, "warning: mocktcp told to send EOF\n");
		flags &= (~MSG_EOF);
	}

	return mt_send(s, (void*)msg, len, flags);
}

int write(int d, const void *buf, size_t nbytes)
{
	if (rm_is_rm_sock(d)) {
		return mt_send(d, (void*)buf, nbytes, 0);
	}
	else {
		return syscall(SYS_write, d, buf, nbytes);
	}
}

int writev(int d, const struct iovec *iov, int iovcnt)
{
	if (rm_is_rm_sock(d)) {
		int	total, i, res, curr;
		char	*buf;

		/* compute total size of data to write */
		for (i=0, total=0; i < iovcnt; i++)
			total += iov[i].iov_len;

		/* coalesce separate data chunks into one big one */
		buf = malloc(total);
		if (buf == NULL) {
			errno = ENOMEM;
			return -1;
		}
		for (i=0, curr=0; i < iovcnt; i++) {
			if (iov[i].iov_base == NULL) continue;
			bcopy(iov[i].iov_base, buf+curr, iov[i].iov_len);
			curr += iov[i].iov_len;
		}
		res = mt_send(d, buf, total, 0);
		free(buf);
		return res;
	}
	else {
		return syscall(SYS_writev, d, iov, iovcnt);
	}
}

int recv(int s, void *buf, size_t len, int flags)
{
	if (flags & MSG_OOB) {
		fprintf(stderr, "warning: mocktcp told to recv OOB data\n");
		flags &= (~MSG_OOB);
	}
	if (flags & MSG_DONTROUTE) {
		fprintf(stderr, "warning: mocktcp told to recv DONTROUTE\n");
		flags &= (~MSG_DONTROUTE);
	}
	if (flags & MSG_EOR) {
		fprintf(stderr, "warning: mocktcp told to recv EOR\n");
		flags &= (~MSG_EOR);
	}
	if (flags & MSG_EOF) {
		fprintf(stderr, "warning: mocktcp told to recv EOF\n");
		flags &= (~MSG_EOF);
	}

	return mt_recv(s, buf, len, flags);
}

int read(int d, void *buf, size_t nbytes)
{
	if (rm_is_rm_sock(d)) {
		return mt_recv(d, buf, nbytes, 0);
	}
	else {
		return syscall(SYS_read, d, buf, nbytes);
	}
}

int readv(int d, const struct iovec *iov, int iovcnt)
{
	if (rm_is_rm_sock(d)) {
		int	i, res, ready, val = 0;

		res = mt_bytes_readable(d, &ready);
		if (res < 0) return -1;

		for (i=0; i < iovcnt && ready > 0; i++) {
			if (iov->iov_base == NULL) continue;
			res = mt_read(d, iov[i].iov_base, iov[i].iov_len);
			if (res < 0) return -1;
			ready -= res;
			val += res;
		}
		return val;
	}
	else {
		return syscall(SYS_readv, d, iov, iovcnt);
	}
}

int setsockopt(int s, int level, int optname, const void *optval, int optlen)
{
	return 0;
}

/*
int getsockopt(int s, int level, int optname, void *optval, int *optlen)
{
}
*/

int getpeername(int s, struct sockaddr *name, int *namelen)
{
	int	res;

	res = mt_getpeername(s, (struct sockaddr_in*)name);
	if (res >=0)
		*namelen = sizeof(struct sockaddr_in);
	return res;
}

int close(int d)
{
	if (rm_is_rm_sock(d)) {
		return mt_close(d);
	}
	else {
		return syscall(SYS_close, d);
	}
}

int bind(int s, const struct sockaddr *name, int namelen)
{
	if (rm_is_rm_sock(s)) {
		struct sockaddr_in *n = (struct sockaddr_in*) name;
		int	port = ntohs(n->sin_port);
		return mt_bind(s, port);
	}
	else {
		return syscall(SYS_bind, s, name, namelen);
	}
}

int select(int nfds, fd_set *readfds, fd_set *writefds, fd_set *exceptfds,
           struct timeval *timeout)
{
	return mt_select(nfds, readfds, writefds, exceptfds, timeout);
}

int shutdown(int s, int how)
{
	/*
	 * We should probably handle this call somehow, but mocktcp and
	 * rmessages does not currently support shutting down only part
	 * of a full duplex connection. It happens to be enough for the
	 * apache web server to NOP this function.
	 */
	return 0;
}

